from __future__ import division, print_function
import datetime
import time
import numpy as np
import scipy as sp
from scipy import linalg
import scipy.special
import scipy.optimize
from functools import partial
import os


def polar2cart(rho, phi):
    """
    convert from polar to cartesian coordinates

    Parameters
    ----------
    rho: 
        radius
    phi: 
        azimuth
    """
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return x, y



def cov_mtx_est(y_mic):
    """
    estimate covariance matrix

    Parameters
    ----------
    y_mic: 
        received signal (complex based band representation) at microphones
    """
    # Q: total number of microphones
    # num_snapshot: number of snapshots used to estimate the covariance matrix
    '''
    Q, num_snapshot = y_mic.shape
    cov_mtx = np.zeros((Q, Q), dtype=complex, order='F')
    for q in range(Q):
        y_mic_outer = y_mic[q, :]
        for qp in range(Q):
            y_mic_inner = y_mic[qp, :]
            cov_mtx[qp, q] = np.dot(y_mic_outer, y_mic_inner.T.conj())
    '''
    cov_mtx = np.dot(np.conj(y_mic), y_mic.T)
    return cov_mtx / y_mic.shape[1]


def extract_off_diag(mtx):
    """
    extract off diagonal entries in mtx.
    The output vector is order in a column major manner.

    Parameters
    ----------
    mtx: 
        input matrix to extract the off diagonal entries
    """
    # we transpose the matrix because the function np.extract will first flatten the matrix
    # withe ordering convention 'C' instead of 'F'!!
    extract_cond = np.reshape((1 - np.eye(*mtx.shape)).T.astype(bool), (-1, 1), order='F')
    return np.reshape(np.extract(extract_cond, mtx.T), (-1, 1), order='F')


def multiband_cov_mtx_est(y_mic):
    """
    estimate covariance matrix based on the received signals at microphones

    Parameters
    ----------
    y_mic: 
        received signal (complex base-band representation) at microphones
    """
    # Q: total number of microphones
    # num_snapshot: number of snapshots used to estimate the covariance matrix
    # num_bands: number of sub-bands considered
    Q, num_snapshot, num_bands = y_mic.shape
    cov_mtx = np.zeros((Q, Q, num_bands), dtype=complex)
    for band in range(num_bands):
        for q in range(Q):
            y_mic_outer = y_mic[q, :, band]
            for qp in range(Q):
                y_mic_inner = y_mic[qp, :, band]
                cov_mtx[qp, q, band] = np.dot(y_mic_outer, y_mic_inner.T.conj())
    return cov_mtx / num_snapshot


def multiband_extract_off_diag(mtx):
    """
    extract off-diagonal entries in mtx
    The output vector is order in a column major manner

    Parameters
    ----------
    mtx: input matrix to extract the off-diagonal entries
    """
    # we transpose the matrix because the function np.extract will first flatten the matrix
    # withe ordering convention 'C' instead of 'F'!!
    Q = mtx.shape[0]
    num_bands = mtx.shape[2]
    extract_cond = np.reshape((1 - np.eye(Q)).T.astype(bool), (-1, 1), order='F')
    return np.column_stack([np.reshape(np.extract(extract_cond, mtx[:, :, band].T),
                                       (-1, 1), order='F')
                            for band in range(num_bands)])


def mtx_freq2raw(M, p_mic_x, p_mic_y):
    """
    build the matrix that maps the Fourier series to the raw microphone signals

    Parameters
    ----------
    M: 
        the Fourier series expansion is limited from -M to M
    p_mic_x: 
        a vector that contains microphones x coordinates
    p_mic_y: 
        a vector that contains microphones y coordinates
    """
    num_mic = p_mic_x.size
    ms = np.reshape(np.arange(-M, M + 1, step=1), (1, -1), order='F')
    G = np.zeros((num_mic, 2 * M + 1), dtype=complex, order='C')
    count_G = 0
    for q in range(num_mic):
        norm_q = np.sqrt(p_mic_x[q] ** 2 + p_mic_y[q] ** 2)
        phi_q = np.arctan2(p_mic_y[q], p_mic_x[q])
        G[q, :] = (-1j) ** ms * sp.special.jv(ms, norm_q) * \
                  np.exp(1j * ms * phi_q)
    return G


def mtx_freq2visi(M, p_mic_x, p_mic_y):
    """
    build the matrix that maps the Fourier series to the visibility

    Parameters
    ----------
    M: 
        the Fourier series expansion is limited from -M to M
    p_mic_x: 
        a vector that constains microphones x coordinates
    p_mic_y: 
        a vector that constains microphones y coordinates
    """
    num_mic = p_mic_x.size
    ms = np.reshape(np.arange(-M, M + 1, step=1), (1, -1), order='F')
    p_mic_x_outer = p_mic_x[:, np.newaxis]
    p_mic_y_outer = p_mic_y[:, np.newaxis]
    p_mic_x_inner = p_mic_x[np.newaxis, :]
    p_mic_y_inner = p_mic_y[np.newaxis, :]
    extract_cond = np.reshape((1 - np.eye(num_mic)).astype(bool), (-1, 1), order='C')
    baseline_x = np.extract(extract_cond, p_mic_x_outer - p_mic_x_inner)[:, np.newaxis]
    baseline_y = np.extract(extract_cond, p_mic_y_outer - p_mic_y_inner)[:, np.newaxis]
    baseline_norm = np.sqrt(baseline_x * baseline_x + baseline_y * baseline_y)
    return (-1j) ** ms * sp.special.jv(ms, baseline_norm) * \
           np.exp(1j * ms * np.arctan2(baseline_y, baseline_x))


def mtx_fri2signal_ri_multiband(M, p_mic_x_all, p_mic_y_all, D1, D2, aslist=False, signal='visibility'):
    """
    build the matrix that maps the Fourier series to the visibility in terms of
    REAL-VALUED entries only. (matrix size double)

    Parameters
    ----------
    M: 
        the Fourier series expansion is limited from -M to M
    p_mic_x_all: 
        a matrix that contains microphones x coordinates
    p_mic_y_all: 
        a matrix that contains microphones y coordinates
    D1: 
        expansion matrix for the real-part
    D2: 
        expansion matrix for the imaginary-part aslist: whether the linear
        mapping for each subband is returned as a list or a block diagonal
        matrix
    signal: 
        The type of signal considered ('visibility' for covariance matrix, 'raw' for microphone inputs)
    """
    num_bands = p_mic_x_all.shape[1]
    if aslist:
        return [mtx_fri2signal_ri(M, p_mic_x_all[:, band_count],
                                  p_mic_y_all[:, band_count], D1, D2, signal)
                for band_count in range(num_bands)]
    else:
        return linalg.block_diag(*[mtx_fri2signal_ri(M, p_mic_x_all[:, band_count],
                                                     p_mic_y_all[:, band_count], D1, D2, signal=signal)
                                   for band_count in range(num_bands)])


def mtx_fri2signal_ri(M, p_mic_x, p_mic_y, D1, D2, signal='visibility'):
    """
    build the matrix that maps the Fourier series to the visibility in terms of
    REAL-VALUED entries only. (matrix size double)

    Parameters
    ----------
    M: 
        the Fourier series expansion is limited from -M to M
    p_mic_x: 
        a vector that contains microphones x coordinates
    p_mic_y: 
        a vector that contains microphones y coordinates
    D1: 
        expansion matrix for the real-part
    D2: 
        expansion matrix for the imaginary-part
    signal: 
        The type of signal considered ('visibility' for covariance matrix, 'raw' for microphone inputs)
    """

    if signal == 'visibility':
        func = mtx_freq2visi
    elif signal == 'raw':
        func = mtx_freq2raw

    return np.dot(cpx_mtx2real(func(M, p_mic_x, p_mic_y)),
                  linalg.block_diag(D1, D2))


def cpx_mtx2real(mtx):
    """
    extend complex valued matrix to an extended matrix of real values only

    Parameters
    ----------
    mtx: 
        input complex valued matrix
    """
    return np.vstack((np.hstack((mtx.real, -mtx.imag)), np.hstack((mtx.imag, mtx.real))))


def hermitian_expan(half_vec_len):
    """
    expand a real-valued vector to a Hermitian symmetric vector.
    The input vector is a concatenation of the real parts with NON-POSITIVE indices and
    the imaginary parts with STRICTLY-NEGATIVE indices.

    Parameters
    ----------
    half_vec_len: 
        length of the first half vector
    """
    D0 = np.eye(half_vec_len)
    D1 = np.vstack((D0, D0[1:, ::-1]))
    D2 = np.vstack((D0, -D0[1:, ::-1]))
    D2 = D2[:, :-1]
    return D1, D2


def output_shrink(K, L):
    """
    shrink the convolution output to half the size.
    used when both the annihilating filter and the uniform samples of sinusoids satisfy
    Hermitian symmetric.

    Parameters
    ----------
    K: 
        the annihilating filter size: K + 1
    L: 
        length of the (complex-valued) b vector
    """
    out_len = L - K
    if out_len % 2 == 0:
        half_out_len = np.int(out_len / 2.)
        mtx_r = np.hstack((np.eye(half_out_len),
                           np.zeros((half_out_len, half_out_len))))
        mtx_i = mtx_r
    else:
        half_out_len = np.int((out_len + 1) / 2.)
        mtx_r = np.hstack((np.eye(half_out_len),
                           np.zeros((half_out_len, half_out_len - 1))))
        mtx_i = np.hstack((np.eye(half_out_len - 1),
                           np.zeros((half_out_len - 1, half_out_len))))
    return linalg.block_diag(mtx_r, mtx_i)


def coef_expan_mtx(K):
    """
    expansion matrix for an annihilating filter of size K + 1

    Parameters
    ----------
    K: 
        number of Dirac. The filter size is K + 1
    """
    if K % 2 == 0:
        D0 = np.eye(np.int(K / 2. + 1))
        D1 = np.vstack((D0, D0[1:, ::-1]))
        D2 = np.vstack((D0, -D0[1:, ::-1]))[:, :-1]
    else:
        D0 = np.eye(np.int((K + 1) / 2.))
        D1 = np.vstack((D0, D0[:, ::-1]))
        D2 = np.vstack((D0, -D0[:, ::-1]))
    return D1, D2


def Tmtx_ri(b_ri, K, D, L):
    """
    build convolution matrix associated with b_ri

    Parameters
    ----------
    b_ri: 
        a real-valued vector
    K: 
        number of Diracs
    D1: 
        expansion matrix for the real-part
    D2: 
        expansion matrix for the imaginary-part
    """
    b_ri = np.dot(D, b_ri)
    b_r = b_ri[:L]
    b_i = b_ri[L:]
    Tb_r = linalg.toeplitz(b_r[K:], b_r[K::-1])
    Tb_i = linalg.toeplitz(b_i[K:], b_i[K::-1])
    return np.vstack((np.hstack((Tb_r, -Tb_i)), np.hstack((Tb_i, Tb_r))))


def Tmtx_ri_half(b_ri, K, D, L, D_coef):
    ''' Split T matrix in conjugate symmetric representation '''
    return np.dot(Tmtx_ri(b_ri, K, D, L), D_coef)


def Tmtx_ri_half_out_half(b_ri, K, D, L, D_coef, mtx_shrink):
    """
    if both b and annihilation filter coefficients are Hermitian symmetric,
    then the output will also be Hermitian symmetric => the effectively output
    is half the size
    """
    return np.dot(np.dot(mtx_shrink, Tmtx_ri(b_ri, K, D, L)), D_coef)


def Rmtx_ri(coef_ri, K, D, L):
    ''' Split T matrix in rea/imaginary representation '''
    coef_ri = np.squeeze(coef_ri)
    coef_r = coef_ri[:K + 1]
    coef_i = coef_ri[K + 1:]
    R_r = linalg.toeplitz(np.concatenate((np.array([coef_r[-1]]),
                                          np.zeros(L - K - 1))),
                          np.concatenate((coef_r[::-1],
                                          np.zeros(L - K - 1)))
                          )
    R_i = linalg.toeplitz(np.concatenate((np.array([coef_i[-1]]),
                                          np.zeros(L - K - 1))),
                          np.concatenate((coef_i[::-1],
                                          np.zeros(L - K - 1)))
                          )
    return np.dot(np.vstack((np.hstack((R_r, -R_i)), np.hstack((R_i, R_r)))), D)


def Rmtx_ri_half(coef_half, K, D, L, D_coef):
    ''' Split T matrix in rea/imaginary conjugate symmetric representation '''
    return Rmtx_ri(np.dot(D_coef, coef_half), K, D, L)


def Rmtx_ri_half_out_half(coef_half, K, D, L, D_coef, mtx_shrink):
    """
    if both b and annihilation filter coefficients are Hermitian symmetric,
    then the output will also be Hermitian symmetric => the effectively output
    is half the size
    """
    return np.dot(mtx_shrink, Rmtx_ri(np.dot(D_coef, coef_half), K, D, L))


def build_mtx_amp(phi_k, p_mic_x, p_mic_y):
    """
    the matrix that maps Diracs' amplitudes to the visibility

    Parameters
    ----------
    phi_k: 
        Diracs' location (azimuth)
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    """
    xk, yk = polar2cart(1, phi_k[np.newaxis, :])
    num_mic = p_mic_x.size
    p_mic_x_outer = p_mic_x[:, np.newaxis]
    p_mic_y_outer = p_mic_y[:, np.newaxis]
    p_mic_x_inner = p_mic_x[np.newaxis, :]
    p_mic_y_inner = p_mic_y[np.newaxis, :]
    extract_cond = np.reshape((1 - np.eye(num_mic)).astype(bool), (-1, 1), order='C')
    baseline_x = np.extract(extract_cond, p_mic_x_outer - p_mic_x_inner)[:, np.newaxis]
    baseline_y = np.extract(extract_cond, p_mic_y_outer - p_mic_y_inner)[:, np.newaxis]
    return np.exp(-1j * (xk * baseline_x + yk * baseline_y))


def build_mtx_amp_ri(p_mic_x, p_mic_y, phi_k):
    ''' builds real/imaginary amplitude matrix '''
    mtx = build_mtx_amp(phi_k, p_mic_x, p_mic_y)
    return np.vstack((mtx.real, mtx.imag))


def build_mtx_raw_amp(p_mic_x, p_mic_y, phi_k):
    """
    the matrix that maps Diracs' amplitudes to the visibility

    Parameters
    ----------
    phi_k: 
        Diracs' location (azimuth)
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    """
    xk, yk = polar2cart(1, phi_k)
    num_mic = p_mic_x.size
    K = phi_k.size
    mtx = np.zeros((num_mic, K), dtype=complex, order='C')
    for q in range(num_mic):
        mtx[q, :] = np.exp(-1j * (xk * p_mic_x[q] + yk * p_mic_y[q]))

    return mtx


def mtx_updated_G_multiband(phi_recon, M, mtx_amp2visi_ri,
                            mtx_fri2visi_ri, num_bands):
    """
    Update the linear transformation matrix that links the FRI sequence to the
    visibilities by using the reconstructed Dirac locations.

    Parameters
    ----------
    phi_recon: 
        the reconstructed Dirac locations (azimuths)
    M: 
        the Fourier series expansion is between -M to M
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    mtx_fri2visi: 
        the linear mapping from Fourier series to visibilities
    """
    L = 2 * M + 1
    ms_half = np.reshape(np.arange(-M, 1, step=1), (-1, 1), order='F')
    phi_recon = np.reshape(phi_recon, (1, -1), order='F')
    mtx_amp2freq = np.exp(-1j * ms_half * phi_recon)  # size: (M + 1) x K
    mtx_amp2freq_ri = np.vstack((mtx_amp2freq.real, mtx_amp2freq.imag[:-1, :]))  # size: (2M + 1) x K
    mtx_fri2amp_ri = linalg.lstsq(mtx_amp2freq_ri, np.eye(L), rcond=None)[0]
    # projection mtx_freq2visi to the null space of mtx_fri2amp
    mtx_null_proj = np.eye(L) - np.dot(mtx_fri2amp_ri.T,
                                       linalg.lstsq(mtx_fri2amp_ri.T, np.eye(L))[0])
    G_updated = np.dot(mtx_amp2visi_ri,
                       linalg.block_diag(*([mtx_fri2amp_ri] * num_bands))
                       ) + \
                np.dot(mtx_fri2visi_ri,
                       linalg.block_diag(*([mtx_null_proj] * num_bands))
                       )
    return G_updated


def mtx_updated_G_multiband_new(phi_opt, M, p_x, p_y,
                                G0_lst, num_bands):
    """
    Update the linear transformation matrix that links the FRI sequence to the
    visibilities by using the reconstructed Dirac locations.

    Parameters
    ----------
    phi_opt: 
        the reconstructed Dirac locations (azimuths)
    M: 
        the Fourier series expansion is between -M to M
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    G0_lst: 
        the original linear mapping from Fourier series to visibilities
    num_bands: 
        number of subbands
    """
    L = 2 * M + 1
    ms_half = np.reshape(np.arange(-M, 1, step=1), (-1, 1), order='F')

    mtx_amp2freq = np.exp(-1j * ms_half * np.reshape(phi_opt, (1, -1), order='F'))
    mtx_amp2freq_ri = np.vstack((mtx_amp2freq.real, mtx_amp2freq.imag[:-1, :]))
    mtx_fri2amp_ri = linalg.lstsq(mtx_amp2freq_ri, np.eye(L))[0]

    G_updated = []
    for band_count in range(num_bands):
        G0_loop = G0_lst[band_count]
        amp_mtx_ri_loop = \
            build_mtx_amp_ri(p_x[:, band_count],
                             p_y[:, band_count],
                             phi_opt)

        high_freq_mapping = np.dot(amp_mtx_ri_loop, mtx_fri2amp_ri)

        # G_updated.append(
        #     G0_loop +
        #     high_freq_mapping -
        #     np.dot(
        #         G0_loop,
        #         linalg.solve(np.dot(G0_loop.T, G0_loop),
        #                      np.dot(G0_loop.T, high_freq_mapping))
        #     )
        # )
        G_updated.append(
            high_freq_mapping +
            G0_loop -
            np.dot(np.dot(G0_loop, mtx_fri2amp_ri.T),
                   linalg.solve(np.dot(mtx_fri2amp_ri, mtx_fri2amp_ri.T),
                                mtx_fri2amp_ri)
                   )
        )

    return G_updated


def mtx_updated_G(phi_recon, M, mtx_amp2visi_ri, mtx_fri2visi_ri):
    """
    Update the linear transformation matrix that links the FRI sequence to the
    visibilities by using the reconstructed Dirac locations.
    Parameters
    ----------
    phi_recon: 
        the reconstructed Dirac locations (azimuths)
    M: 
        the Fourier series expansion is between -M to M
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    mtx_freq2visi: 
        the linear mapping from Fourier series to visibilities
    """
    L = 2 * M + 1
    ms_half = np.reshape(np.arange(-M, 1, step=1), (-1, 1), order='F')
    phi_recon = np.reshape(phi_recon, (1, -1), order='F')
    mtx_amp2freq = np.exp(-1j * ms_half * phi_recon)  # size: (M + 1) x K
    mtx_amp2freq_ri = np.vstack((mtx_amp2freq.real, mtx_amp2freq.imag[:-1, :]))  # size: (2M + 1) x K
    mtx_fri2amp_ri = linalg.lstsq(mtx_amp2freq_ri, np.eye(L))[0]
    # projection mtx_freq2visi to the null space of mtx_fri2amp
    mtx_null_proj = np.eye(L) - np.dot(mtx_fri2amp_ri.T,
                                       linalg.lstsq(mtx_fri2amp_ri.T, np.eye(L))[0])
    G_updated = np.dot(mtx_amp2visi_ri, mtx_fri2amp_ri) + \
                np.dot(mtx_fri2visi_ri, mtx_null_proj)
    return G_updated


def dirac_recon_ri(G, a_ri, K, M, noise_level, max_ini=100, stop_cri='mse'):
    """
    Reconstruct point sources' locations (azimuth) from the visibility measurements

    Parameters
    ----------
    G: 
        the linear transformation matrix that links the visibilities to
        uniformly sampled sinusoids
    a_ri: 
        the visibility measurements
    K: 
        number of Diracs
    M: 
        the Fourier series expansion is between -M and M
    noise_level: 
        level of noise (ell_2 norm) in the measurements
    max_ini: 
        maximum number of initialisations
    stop_cri: 
        stopping criterion, either 'mse' or 'max_iter'
    """
    L = 2 * M + 1  # length of the (complex-valued) b vector
    a_ri = a_ri.flatten('F')
    compute_mse = (stop_cri == 'mse')
    # size of G: (Q(Q-1)) x (2M + 1), where Q is the number of antennas
    assert not np.iscomplexobj(G)  # G should be real-valued
    GtG = np.dot(G.T, G)
    Gt_a = np.dot(G.T, a_ri)
    # maximum number of iterations with each initialisation
    max_iter = 50
    min_error = float('inf')
    # the least-square solution
    beta_ri = linalg.lstsq(G, a_ri)[0]
    D1, D2 = hermitian_expan(M + 1)
    D = linalg.block_diag(D1, D2)
    # size of Tbeta_ri: 2(L - K) x 2(K + 1)
    Tbeta_ri = Tmtx_ri(beta_ri, K, D, L)

    # size of various matrices / vectors
    sz_G1 = L

    sz_Tb0 = 2 * (L - K)
    sz_Tb1 = 2 * (K + 1)

    sz_Rc0 = 2 * (L - K)
    sz_Rc1 = L

    sz_coef = 2 * (K + 1)
    sz_bri = L

    rhs = np.append(np.zeros(sz_coef + sz_Tb0 + sz_G1, dtype=float), 1)
    rhs_bl = np.concatenate((Gt_a, np.zeros(sz_Rc0, dtype=Gt_a.dtype)))

    # the main iteration with different random initialisations
    for ini in range(max_ini):
        c_ri = np.random.randn(sz_coef, 1)
        c0_ri = c_ri.copy()
        error_seq = np.zeros(max_iter, dtype=float)
        R_loop = Rmtx_ri(c_ri, K, D, L)
        for inner in range(max_iter):
            mtx_loop = np.vstack((np.hstack((np.zeros((sz_coef, sz_coef)),
                                             Tbeta_ri.T,
                                             np.zeros((sz_coef, sz_Rc1)),
                                             c0_ri)),
                                  np.hstack((Tbeta_ri,
                                             np.zeros((sz_Tb0, sz_Tb0)),
                                             -R_loop,
                                             np.zeros((sz_Rc0, 1))
                                             )),
                                  np.hstack((np.zeros((sz_Rc1, sz_Tb1)),
                                             -R_loop.T,
                                             GtG,
                                             np.zeros((sz_Rc1, 1))
                                             )),
                                  np.hstack((c0_ri.T,
                                             np.zeros((1, sz_Tb0 + sz_Rc1 + 1))
                                             ))
                                  ))
            # matrix should be symmetric
            mtx_loop = (mtx_loop + mtx_loop.T) / 2.
            c_ri = linalg.lstsq(mtx_loop, rhs)[0][:sz_coef]
            # c_ri = linalg.solve(mtx_loop, rhs)[:sz_coef]

            R_loop = Rmtx_ri(c_ri, K, D, L)
            mtx_brecon = np.vstack((np.hstack((GtG, R_loop.T)),
                                    np.hstack((R_loop, np.zeros((sz_Rc0, sz_Rc0))))
                                    ))
            mtx_brecon = (mtx_brecon + mtx_brecon.T) / 2.
            b_recon_ri = linalg.lstsq(mtx_brecon, rhs_bl)[0][:sz_bri]
            # b_recon_ri = linalg.solve(mtx_brecon, rhs_bl)[:sz_bri]

            error_seq[inner] = linalg.norm(a_ri - np.dot(G, b_recon_ri))
            if error_seq[inner] < min_error:
                min_error = error_seq[inner]
                b_opt = np.dot(D1, b_recon_ri[:M + 1]) + 1j * np.dot(D2, b_recon_ri[M + 1:])
                c_opt = c_ri[:K + 1] + 1j * c_ri[K + 1:]  # real and imaginary parts

            if min_error < noise_level and compute_mse:
                break

        if min_error < noise_level and compute_mse:
            break
    return c_opt, min_error, b_opt, ini


def dirac_recon_ri_half(G, a_ri, K, M, noise_level, max_ini=100, stop_cri='mse'):
    """
    Reconstruct point sources' locations (azimuth) from the visibility measurements.
    Here we enforce hermitian symmetry in the annihilating filter coefficients so that
    roots on the unit circle are encouraged.

    Parameters
    ----------
    param G: 
        the linear transformation matrix that links the visibilities to
        uniformly sampled sinusoids
    a_ri: 
        the visibility measurements
    K: 
        number of Diracs
    M: 
        the Fourier series expansion is between -M and M
    noise_level: 
        level of noise (ell_2 norm) in the measurements
    max_ini: 
        maximum number of initialisations
    stop_cri: 
        stopping criterion, either 'mse' or 'max_iter'
    """
    L = 2 * M + 1  # length of the (complex-valued) b vector
    a_ri = a_ri.flatten('F')
    compute_mse = (stop_cri == 'mse')
    # size of G: (Q(Q-1)) x (2M + 1), where Q is the number of antennas
    assert not np.iscomplexobj(G)  # G should be real-valued
    GtG = np.dot(G.T, G)
    Gt_a = np.dot(G.T, a_ri)
    # maximum number of iterations with each initialisation
    max_iter = 50
    min_error = float('inf')
    # the least-square solution
    beta_ri = linalg.lstsq(G, a_ri)[0]
    D1, D2 = hermitian_expan(M + 1)
    D = linalg.block_diag(D1, D2)
    D_coef1, D_coef2 = coef_expan_mtx(K)
    D_coef = linalg.block_diag(D_coef1, D_coef2)
    # size of Tbeta_ri: 2(L - K) x 2(K + 1)
    Tbeta_ri = Tmtx_ri_half(beta_ri, K, D, L, D_coef)

    # size of various matrices / vectors
    sz_G1 = L

    sz_Tb0 = 2 * (L - K)
    sz_Tb1 = K + 1

    sz_Rc0 = 2 * (L - K)
    sz_Rc1 = L

    sz_coef = K + 1
    sz_bri = L

    rhs = np.append(np.zeros(sz_coef + sz_Tb0 + sz_G1, dtype=float), 1)
    rhs_bl = np.concatenate((Gt_a, np.zeros(sz_Rc0, dtype=Gt_a.dtype)))

    # the main iteration with different random initialisations
    for ini in range(max_ini):
        c_ri_half = np.random.randn(sz_coef, 1)
        c0_ri_half = c_ri_half.copy()
        error_seq = np.zeros(max_iter, dtype=float)
        R_loop = Rmtx_ri_half(c_ri_half, K, D, L, D_coef)
        for inner in range(max_iter):
            mtx_loop = np.vstack((np.hstack((np.zeros((sz_coef, sz_coef)),
                                             Tbeta_ri.T,
                                             np.zeros((sz_coef, sz_Rc1)),
                                             c0_ri_half)),
                                  np.hstack((Tbeta_ri,
                                             np.zeros((sz_Tb0, sz_Tb0)),
                                             -R_loop,
                                             np.zeros((sz_Rc0, 1))
                                             )),
                                  np.hstack((np.zeros((sz_Rc1, sz_Tb1)),
                                             -R_loop.T,
                                             GtG,
                                             np.zeros((sz_Rc1, 1))
                                             )),
                                  np.hstack((c0_ri_half.T,
                                             np.zeros((1, sz_Tb0 + sz_Rc1 + 1))
                                             ))
                                  ))
            # matrix should be symmetric
            mtx_loop = (mtx_loop + mtx_loop.T) / 2.
            c_ri_half = linalg.lstsq(mtx_loop, rhs)[0][:sz_coef]
            # c_ri = linalg.solve(mtx_loop, rhs)[:sz_coef]

            R_loop = Rmtx_ri_half(c_ri_half, K, D, L, D_coef)
            mtx_brecon = np.vstack((np.hstack((GtG, R_loop.T)),
                                    np.hstack((R_loop, np.zeros((sz_Rc0, sz_Rc0))))
                                    ))
            mtx_brecon = (mtx_brecon + mtx_brecon.T) / 2.
            b_recon_ri = linalg.lstsq(mtx_brecon, rhs_bl)[0][:sz_bri]
            # b_recon_ri = linalg.solve(mtx_brecon, rhs_bl)[:sz_bri]

            error_seq[inner] = linalg.norm(a_ri - np.dot(G, b_recon_ri))
            if error_seq[inner] < min_error:
                min_error = error_seq[inner]
                b_opt = np.dot(D1, b_recon_ri[:M + 1]) + 1j * np.dot(D2, b_recon_ri[M + 1:])
                c_ri = np.dot(D_coef, c_ri_half)
                c_opt = c_ri[:K + 1] + 1j * c_ri[K + 1:]  # real and imaginary parts

            if min_error < noise_level and compute_mse:
                break

        if min_error < noise_level and compute_mse:
            break
    return c_opt, min_error, b_opt, ini

def dirac_recon_ri_half_multiband_lu(G_lst, GtG_lst, GtG_inv_lst, a_ri, K, M, max_ini=100, max_iter=50):
    """
    Here we use LU decomposition to precompute a few entries.  Reconstruct
    point sources' locations (azimuth) from the visibility measurements.  Here
    we enforce hermitian symmetry in the annihilating filter coefficients so
    that roots on the unit circle are encouraged.

    Parameters
    ----------
    G_lst: 
        a list of the linear transformation matrices that links the
        visibilities to uniformly sampled sinusoids
    a_ri: 
        the visibility measurements
    K: 
        number of Diracs
    M: 
        the Fourier series expansion is between -M and M
    noise_level: 
        level of noise (ell_2 norm) in the measurements
    max_ini: 
        maximum number of initialisations
    stop_cri: 
        stopping criterion, either 'mse' or 'max_iter'
    """
    num_bands = a_ri.shape[1]  # number of bands considered
    L = 2 * M + 1  # length of the (complex-valued) b vector for each band
    assert not np.iscomplexobj(np.concatenate(G_lst))  # G should be real-valued
    assert not np.iscomplexobj(np.concatenate(a_ri))  # a_ri should be real-valued

    D1, D2 = hermitian_expan(M + 1)
    D = linalg.block_diag(D1, D2)
    D_coef1, D_coef2 = coef_expan_mtx(K)
    D_coef = linalg.block_diag(D_coef1, D_coef2)

    # shrink the output size to half as both the annihilating filter coeffiicnets and
    # the uniform samples of sinusoids are Hermitian symmetric
    mtx_shrink = output_shrink(K, L)

    # compute Tbeta and GtG for each subband
    Gt_a_lst = []
    beta_ri_lst = []
    Tbeta_ri_lst = []
    for loop in range(num_bands):
        #G_loop = G_lst[loop]
        #a_loop = a_ri[:, loop]

        #lu_GtG_loop = linalg.lu_factor(np.dot(G_loop.T, G_loop), check_finite=False)

        Gt_a_loop = np.dot(G_lst[loop].T, a_ri[:, loop])

        #beta_ri_loop = linalg.lu_solve(lu_GtG_loop, Gt_a_loop, check_finite=False)
        beta_ri_loop = np.dot(GtG_inv_lst[loop], Gt_a_loop)

        beta_ri_lst.append(beta_ri_loop)
        Gt_a_lst.append(Gt_a_loop)

        Tbeta_ri_lst.append(Tmtx_ri_half_out_half(
            beta_ri_loop, K, D, L, D_coef, mtx_shrink))

    sz_coef = K + 1
    rhs = np.append(np.zeros(sz_coef, dtype=float), 1)
    min_error = float('inf')

    for ini in range(max_ini):
        c_ri_half = np.random.randn(sz_coef, 1)
        c0_ri_half = c_ri_half.copy()
        Rmtx_band = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)

        mtx_loop = \
            np.vstack((
                np.hstack((
                    lu_compute_mtx_obj_initial(GtG_inv_lst, Tbeta_ri_lst,
                                               Rmtx_band, num_bands, K),
                    c0_ri_half)),
                np.append(c0_ri_half, 0).T
            ))
        for inner in range(max_iter):
            # update mtx_loop
            if inner != 0:
                # mtx_loop[:sz_coef, :sz_coef] = \
                #     lu_compute_mtx_obj(Tbeta_ri_lst, num_bands, K, lu_R_GtGinv_Rt_loop)
                mtx_loop[:sz_coef, :sz_coef] = mtx_loop_upper_left
            try:
                c_ri_half = linalg.solve(mtx_loop, rhs, check_finite=False)[:sz_coef]
            except linalg.LinAlgError:
                break

            # build R based on the updated c
            Rmtx_band = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)

            # update b_recon
            error_loop, mtx_loop_upper_left = \
                compute_obj_val(GtG_inv_lst, Tbeta_ri_lst, Rmtx_band, c_ri_half, num_bands, K)

            if error_loop < min_error:
                min_error = error_loop
                # b_opt = np.dot(D1, b_recon_ri[:M + 1, :]) + \
                #         1j * np.dot(D2, b_recon_ri[M + 1:, :])
                Rmtx_opt = Rmtx_band
                c_ri = np.dot(D_coef, c_ri_half)
                c_opt = c_ri[:K + 1] + 1j * c_ri[K + 1:]  # real and imaginary parts

    b_opt_ri = compute_b(G_lst, GtG_lst, beta_ri_lst, Rmtx_opt, num_bands, a_ri, use_lu=True, GtG_inv_lst=GtG_inv_lst)[0]
    b_opt = np.dot(D1, b_opt_ri[:M + 1, :]) + 1j * np.dot(D2, b_opt_ri[M + 1:, :])

    return c_opt, min_error, b_opt


def dirac_recon_ri_half_multiband(G_lst, a_ri, K, M, max_ini=100):
    """
    Reconstruct point sources' locations (azimuth) from the visibility measurements.
    Here we enforce hermitian symmetry in the annihilating filter coefficients so that
    roots on the unit circle are encouraged.

    Parameters
    ----------
    G_lst: 
        a list of the linear transformation matrices that links the
        visibilities to uniformly sampled sinusoids
    a_ri: 
        the visibility measurements
    K: 
        number of Diracs
    M: 
        the Fourier series expansion is between -M and M
    noise_level: 
        level of noise (ell_2 norm) in the measurements
    max_ini: 
        maximum number of initialisations
    stop_cri: 
        stopping criterion, either 'mse' or 'max_iter'
    """
    num_bands = a_ri.shape[1]  # number of bands considered
    L = 2 * M + 1  # length of the (complex-valued) b vector for each band
    assert not np.iscomplexobj(np.concatenate(G_lst))  # G should be real-valued
    assert not np.iscomplexobj(np.concatenate(a_ri))  # a_ri should be real-valued

    # maximum number of iterations with each initialisation
    max_iter = 50

    D1, D2 = hermitian_expan(M + 1)
    D = linalg.block_diag(D1, D2)
    D_coef1, D_coef2 = coef_expan_mtx(K)
    D_coef = linalg.block_diag(D_coef1, D_coef2)

    # shrink the output size to half as both the annihilating filter coeffiicnets and
    # the uniform samples of sinusoids are Hermitian symmetric
    mtx_shrink = output_shrink(K, L)

    # compute Tbeta and GtG for each subband
    Gt_a_lst = []
    GtG_lst = []
    beta_ri_lst = []
    Tbeta_ri_lst = []
    for loop in range(num_bands):
        G_loop = G_lst[loop]
        a_loop = a_ri[:, loop]
        beta_ri_loop = linalg.lstsq(G_loop, a_loop)[0]

        beta_ri_lst.append(beta_ri_loop)
        Gt_a_lst.append(np.dot(G_loop.T, a_loop))
        GtG_lst.append(np.dot(G_loop.T, G_loop))

        Tbeta_ri_lst.append(Tmtx_ri_half_out_half(
            beta_ri_loop, K, D, L, D_coef, mtx_shrink))

    sz_coef = K + 1
    rhs = np.append(np.zeros(sz_coef, dtype=float), 1)
    min_error = float('inf')

    for ini in range(max_ini):
        c_ri_half = np.random.randn(sz_coef, 1)
        c0_ri_half = c_ri_half.copy()
        Rmtx_band = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)
        mtx_loop = np.vstack((np.hstack((compute_mtx_obj(GtG_lst, Tbeta_ri_lst,
                                                         Rmtx_band, num_bands, K),
                                         c0_ri_half)),
                              np.append(c0_ri_half, 0).T
                              ))
        for inner in range(max_iter):
            # update mtx_loop
            if inner != 0:
                mtx_loop[:sz_coef, :sz_coef] = \
                    compute_mtx_obj(GtG_lst, Tbeta_ri_lst, Rmtx_band, num_bands, K)

            # matrix should be symmetric
            mtx_loop += mtx_loop.T
            mtx_loop *= 0.5
            c_ri_half = linalg.solve(mtx_loop, rhs, check_finite=False)[:sz_coef]

            # build R based on the updated c
            Rmtx_band = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)

            # update b_recon
            b_recon_ri, error_loop = \
                compute_b(G_lst, GtG_lst, beta_ri_lst, Rmtx_band, num_bands, a_ri)

            if error_loop < min_error:
                min_error = error_loop
                b_opt = np.dot(D1, b_recon_ri[:M + 1, :]) + \
                        1j * np.dot(D2, b_recon_ri[M + 1:, :])
                c_ri = np.dot(D_coef, c_ri_half)
                c_opt = c_ri[:K + 1] + 1j * c_ri[K + 1:]  # real and imaginary parts
    return c_opt, min_error, b_opt


def lu_compute_mtx_obj(Tbeta_lst, num_bands, K, lu_R_GtGinv_Rt_lst):
    """
    compute the matrix (M) in the objective function:
        min   c^H M c
        s.t.  c0^H c = 1

    Parameters
    ----------
    GtG_lst: 
        list of G^H * G
    Tbeta_lst: 
        list of Teoplitz matrices for beta-s
    Rc0: 
        right dual matrix for the annihilating filter (same of each block -> not a list)
    """
    mtx = np.zeros((K + 1, K + 1), dtype=float)  # <= assume G, Tbeta and Rc0 are real-valued

    for loop in range(num_bands):
        Tbeta_loop = Tbeta_lst[loop]
        mtx += np.dot(Tbeta_loop.T,
                      linalg.lu_solve(lu_R_GtGinv_Rt_lst[loop],
                                      Tbeta_loop, check_finite=False)
                      )

    return mtx


def lu_compute_mtx_obj_initial(GtG_inv_lst, Tbeta_lst, Rc0, num_bands, K):
    """
    compute the matrix (M) in the objective function:
        min   c^H M c
        s.t.  c0^H c = 1

    Parameters
    ----------
    GtG_lst: 
        list of G^H * G
    Tbeta_lst: 
        list of Teoplitz matrices for beta-s
    Rc0: 
        right dual matrix for the annihilating filter (same of each block -> not a list)
    """
    mtx = np.zeros((K + 1, K + 1), dtype=float)  # <= assume G, Tbeta and Rc0 are real-valued
    for loop in range(num_bands):
        Tbeta_loop = Tbeta_lst[loop]
        mtx += np.dot(Tbeta_loop.T,
                      linalg.solve(np.dot(Rc0,
                                          np.dot(GtG_inv_lst[loop], Rc0.T)),
                                   Tbeta_loop)
                      )
    return mtx


def compute_mtx_obj(GtG_lst, Tbeta_lst, Rc0, num_bands, K):
    """
    compute the matrix (M) in the objective function:
        min   c^H M c
        s.t.  c0^H c = 1

    Parameters
    ----------
    GtG_lst: 
        list of G^H * G
    Tbeta_lst: 
        list of Teoplitz matrices for beta-s
    Rc0: 
        right dual matrix for the annihilating filter (same of each block -> not a list)
    """
    mtx = np.zeros((K + 1, K + 1), dtype=float)  # <= assume G, Tbeta and Rc0 are real-valued
    for loop in range(num_bands):
        Tbeta_loop = Tbeta_lst[loop]
        GtG_loop = GtG_lst[loop]
        mtx += np.dot(Tbeta_loop.T,
                      linalg.solve(np.dot(Rc0, linalg.solve(GtG_loop, Rc0.T)),
                                   Tbeta_loop)
                      )
    return mtx

def compute_obj_val(GtG_inv_lst, Tbeta_lst, Rc0, c_ri_half, num_bands, K):
    """
    compute the fitting error.
    CAUTION: Here we assume use_lu = True
    """
    mtx = np.zeros((K + 1, K + 1), dtype=float)  # <= assume G, Tbeta and Rc0 are real-valued
    fitting_error = 0
    for band_count in range(num_bands):
        #GtG_loop = GtG_lst[band_count]
        Tbeta_loop = Tbeta_lst[band_count]

        mtx += np.dot(Tbeta_loop.T,
                      linalg.solve(
                          np.dot(Rc0, 
                              #linalg.lu_solve(GtG_loop, Rc0.T, check_finite=False)
                              np.dot(GtG_inv_lst[band_count], Rc0.T)
                              ),
                          Tbeta_loop, check_finite=False
                      )
                      )

    fitting_error += np.sqrt(np.dot(c_ri_half.T, np.dot(mtx, c_ri_half)).real)

    return fitting_error, mtx


def compute_b(G_lst, GtG_lst, beta_lst, Rc0, num_bands, a_ri, use_lu=False, GtG_inv_lst=None):
    """
    compute the uniform sinusoidal samples b from the updated annihilating
    filter coeffiients.

    Parameters
    ----------
    GtG_lst: 
        list of G^H G for different subbands
    beta_lst: 
        list of beta-s for different subbands
    Rc0: 
        right-dual matrix, here it is the convolution matrix associated with c
    num_bands: 
        number of bands
    a_ri: 
        a 2D numpy array. each column corresponds to the measurements within a subband
    """
    b_lst = []
    a_Gb_lst = []
    if use_lu:
        assert GtG_inv_lst is not None
        lu_lst = []
        for loop in range(num_bands):
            GtG_loop = GtG_lst[loop]
            beta_loop = beta_lst[loop]
            lu_loop = linalg.lu_factor(
                np.dot(Rc0, np.dot(GtG_inv_lst[loop], Rc0.T)),
                check_finite=False
            )
            b_loop = beta_loop - \
                     np.dot(GtG_inv_lst[loop],
                                     np.dot(Rc0.T,
                                            linalg.lu_solve(lu_loop, np.dot(Rc0, beta_loop),
                                                            check_finite=False)),
                                     )

            b_lst.append(b_loop)
            a_Gb_lst.append(a_ri[:, loop] - np.dot(G_lst[loop], b_loop))
            lu_lst.append(lu_loop)

        return np.column_stack(b_lst), linalg.norm(np.concatenate(a_Gb_lst)), lu_lst
    else:
        for loop in range(num_bands):
            GtG_loop = GtG_lst[loop]
            beta_loop = beta_lst[loop]
            b_loop = beta_loop - \
                     linalg.solve(GtG_loop,
                                  np.dot(Rc0.T,
                                         linalg.solve(np.dot(Rc0, linalg.solve(GtG_loop, Rc0.T)),
                                                      np.dot(Rc0, beta_loop)))
                                  )

            b_lst.append(b_loop)
            a_Gb_lst.append(a_ri[:, loop] - np.dot(G_lst[loop], b_loop))

        return np.column_stack(b_lst), linalg.norm(np.concatenate(a_Gb_lst))


def dirac_recon_ri_half_multiband_parallel(G, a_ri, K, M, max_ini=100):
    """
    Reconstruct point sources' locations (azimuth) from the visibility measurements.
    Here we enforce hermitian symmetry in the annihilating filter coefficients so that
    roots on the unit circle are encouraged.
    We use parallel implementation when stop_cri == 'max_iter'

    Parameters
    ----------
    G:
        the linear transformation matrix that links the visibilities to
        uniformly sampled sinusoids
    a_ri: 
        the visibility measurements
    K: 
        number of Diracs
    M: 
        the Fourier series expansion is between -M and M
    noise_level: 
        level of noise (ell_2 norm) in the measurements
    max_ini: 
        maximum number of initialisations
    stop_cri: 
        stopping criterion, either 'mse' or 'max_iter'
    """
    num_bands = a_ri.shape[1]  # number of bands considered
    L = 2 * M + 1  # length of the (complex-valued) b vector for each band
    a_ri = a_ri.flatten('F')
    # size of G: (Q(Q-1)) x (2M + 1), where Q is the number of antennas
    assert not np.iscomplexobj(G)  # G should be real-valued
    Gt_a = np.dot(G.T, a_ri)
    GtG = np.dot(G.T, G)
    # maximum number of iterations with each initialisation
    max_iter = 50

    # the least-square solution
    beta_ri = np.reshape(linalg.lstsq(G, a_ri)[0], (-1, num_bands), order='F')
    D1, D2 = hermitian_expan(M + 1)
    D = linalg.block_diag(D1, D2)
    D_coef1, D_coef2 = coef_expan_mtx(K)
    D_coef = linalg.block_diag(D_coef1, D_coef2)

    # shrink the output size to half as both the annihilating filter coeffiicnets and
    # the uniform samples of sinusoids are Hermitian symmetric
    mtx_shrink = output_shrink(K, L)

    # size of Tbeta_ri: (L - K)num_bands x 2(K + 1)
    Tbeta_ri = np.vstack([Tmtx_ri_half_out_half(beta_ri[:, band_count],
                                                K, D, L, D_coef, mtx_shrink)
                          for band_count in range(num_bands)])

    # size of various matrices / vectors
    sz_G1 = L * num_bands

    sz_Tb0 = (L - K) * num_bands  # the only effective size because of Hermitian symmetry

    sz_Rc0 = (L - K) * num_bands

    sz_coef = K + 1

    rhs = np.append(np.zeros(sz_coef + sz_Tb0 + sz_G1, dtype=float), 1)
    rhs_bl = np.concatenate((Gt_a, np.zeros(sz_Rc0, dtype=float)))

    # the main iteration with different random initialisations
    partial_dirac_recon = partial(dirac_recon_ri_multiband_inner,
                                  a_ri=a_ri, num_bands=num_bands, rhs=rhs,
                                  rhs_bl=rhs_bl, K=K, M=M, D1=D1, D2=D2,
                                  D_coef=D_coef, mtx_shrink=mtx_shrink,
                                  Tbeta_ri=Tbeta_ri, G=G, GtG=GtG, max_iter=max_iter)

    # generate all the random initialisations
    c_ri_half_all = np.random.randn(sz_coef, max_ini)

    res_all = []
    for loop in range(max_ini):
        res_all.append(
            partial_dirac_recon(c_ri_half_all[:, loop][:, np.newaxis])
        )

    # find the one with smallest error
    min_idx = np.array(zip(*res_all)[1]).argmin()
    c_opt, min_error, b_opt = res_all[min_idx]

    return c_opt, min_error, b_opt


def dirac_recon_ri_multiband_inner(c_ri_half, a_ri, num_bands, rhs, rhs_bl, K, M,
                                   D1, D2, D_coef, mtx_shrink, Tbeta_ri,
                                   G, GtG, max_iter):
    ''' Inner loop of the `dirac_recon_ri_multiband` function '''
    min_error = float('inf')
    # size of various matrices / vectors
    L = 2 * M + 1  # length of the (complex-valued) b vector

    sz_Tb0 = (L - K) * num_bands  # the only effective size because of Hermitian symmetry
    sz_Tb1 = K + 1

    sz_Rc0 = (L - K) * num_bands
    sz_Rc1 = L * num_bands

    sz_coef = K + 1
    sz_bri = L * num_bands

    # indices where the 4 x 4 block matrix is updated at each iteration
    # for -R(c)
    row_s1, row_e1 = sz_coef, sz_coef + sz_Tb0
    col_s1, col_e1 = sz_Tb0 + sz_Tb1, sz_Tb0 + sz_Tb1 + sz_Rc1
    # for -R(c).T
    row_s2, row_e2 = sz_coef + sz_Tb0, sz_coef + sz_Tb0 + sz_Rc1
    col_s2, col_e2 = sz_coef, sz_coef + sz_Tb0

    # GtG = np.dot(G.T, G)
    D = linalg.block_diag(D1, D2)

    c0_ri_half = c_ri_half.copy()
    error_seq = np.zeros(max_iter, dtype=float)

    Rmtx_band = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)
    R_loop = linalg.block_diag(*([Rmtx_band] * num_bands))

    # first row of mtx_loop
    mtx_loop_first_row = np.hstack((np.zeros((sz_coef, sz_coef)), Tbeta_ri.T,
                                    np.zeros((sz_coef, sz_Rc1)), c0_ri_half))
    # last row of mtx_loop
    mtx_loop_last_row = np.hstack((c0_ri_half.T, np.zeros((1, sz_Tb0 + sz_Rc1 + 1))))

    # initialise mtx_loop and mtx_brecon
    mtx_loop = np.vstack((mtx_loop_first_row,
                          np.hstack((Tbeta_ri,
                                     np.zeros((sz_Tb0, sz_Tb0)),
                                     -R_loop,
                                     np.zeros((sz_Rc0, 1))
                                     )),
                          np.hstack((np.zeros((sz_Rc1, sz_Tb1)),
                                     -R_loop.T,
                                     GtG,
                                     np.zeros((sz_Rc1, 1))
                                     )),
                          mtx_loop_last_row
                          ))

    mtx_brecon = np.vstack((np.hstack((GtG, R_loop.T)),
                            np.hstack((R_loop, np.zeros((sz_Rc0, sz_Rc0))))
                            ))

    for inner in range(max_iter):
        # update the mtx_loop matrix
        mtx_loop[row_s1:row_e1, col_s1:col_e1] = -R_loop
        mtx_loop[row_s2:row_e2, col_s2:col_e2] = -R_loop.T

        # matrix should be symmetric
        mtx_loop += mtx_loop.T
        mtx_loop *= 0.5
        c_ri_half = linalg.solve(mtx_loop, rhs, check_finite=False)[:sz_coef]

        Rmtx_band = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)
        R_loop = linalg.block_diag(*([Rmtx_band] * num_bands))

        # update mtx_brecon
        mtx_brecon[:sz_Rc1, sz_Rc1:] = R_loop.T
        mtx_brecon[sz_Rc1:, :sz_Rc1] = R_loop

        mtx_brecon += mtx_brecon.T
        mtx_brecon *= 0.5
        b_recon_ri = linalg.solve(mtx_brecon, rhs_bl, check_finite=False)[:sz_bri]

        error_seq[inner] = linalg.norm(a_ri - np.dot(G, b_recon_ri))
        if error_seq[inner] < min_error:
            min_error = error_seq[inner]
            b_recon_ri = np.reshape(b_recon_ri, (-1, num_bands), order='F')
            b_opt = np.dot(D1, b_recon_ri[:M + 1, :]) + \
                    1j * np.dot(D2, b_recon_ri[M + 1:, :])
            c_ri = np.dot(D_coef, c_ri_half)
            c_opt = c_ri[:K + 1] + 1j * c_ri[K + 1:]  # real and imaginary parts
    return c_opt, min_error, b_opt


def dirac_recon_ri_half_parallel(G, a_ri, K, M, max_ini=100):
    """
    Reconstruct point sources' locations (azimuth) from the visibility measurements.
    Here we enforce hermitian symmetry in the annihilating filter coefficients so that
    roots on the unit circle are encouraged.
    We use parallel implementation when stop_cri == 'max_iter'

    Parameters
    ----------
    G: 
        the linear transformation matrix that links the visibilities to
        uniformly sampled sinusoids
    a_ri: 
        the visibility measurements
    K: 
        number of Diracs
    M: 
        the Fourier series expansion is between -M and M
    noise_level: 
        level of noise (ell_2 norm) in the measurements
    max_ini: 
        maximum number of initialisations
    stop_cri: 
        stopping criterion, either 'mse' or 'max_iter'
    """
    L = 2 * M + 1  # length of the (complex-valued) b vector
    a_ri = a_ri.flatten('F')
    # size of G: (Q(Q-1)) x (2M + 1), where Q is the number of antennas
    assert not np.iscomplexobj(G)  # G should be real-valued
    Gt_a = np.dot(G.T, a_ri)
    # maximum number of iterations with each initialisation
    max_iter = 50

    # the least-square solution
    beta_ri = linalg.lstsq(G, a_ri)[0]
    D1, D2 = hermitian_expan(M + 1)
    D = linalg.block_diag(D1, D2)
    D_coef1, D_coef2 = coef_expan_mtx(K)
    D_coef = linalg.block_diag(D_coef1, D_coef2)

    # shrink the output size to half as both the annihilating filter coeffiicnets and
    # the uniform samples of sinusoids are Hermitian symmetric
    mtx_shrink = output_shrink(K, L)

    # size of Tbeta_ri: (L - K) x 2(K + 1)
    Tbeta_ri = Tmtx_ri_half_out_half(beta_ri, K, D, L, D_coef, mtx_shrink)

    # size of various matrices / vectors
    sz_G1 = L

    sz_Tb0 = L - K  # the only effective size because of Hermitian symmetry

    sz_Rc0 = L - K

    sz_coef = K + 1

    rhs = np.append(np.zeros(sz_coef + sz_Tb0 + sz_G1, dtype=float), 1)
    rhs_bl = np.concatenate((Gt_a, np.zeros(sz_Rc0, dtype=Gt_a.dtype)))

    # the main iteration with different random initialisations
    partial_dirac_recon = partial(dirac_recon_ri_inner, a_ri=a_ri, rhs=rhs,
                                  rhs_bl=rhs_bl, K=K, M=M, D1=D1, D2=D2,
                                  D_coef=D_coef, mtx_shrink=mtx_shrink,
                                  Tbeta_ri=Tbeta_ri, G=G, max_iter=max_iter)

    # generate all the random initialisations
    c_ri_half_all = np.random.randn(sz_coef, max_ini)

    res_all = []
    for loop in range(max_ini):
        res_all.append(
                partial_dirac_recon(c_ri_half_all[:,loop][:,np.newaxis])
                )

    # find the one with smallest error
    min_idx = np.array(zip(*res_all)[1]).argmin()
    c_opt, min_error, b_opt = res_all[min_idx]

    return c_opt, min_error, b_opt


def dirac_recon_ri_inner(c_ri_half, a_ri, rhs, rhs_bl, K, M,
                         D1, D2, D_coef, mtx_shrink, Tbeta_ri, G, max_iter):
    ''' inner loop of the `dirac_recon_ri_half_parallel` function '''

    min_error = float('inf')
    # size of various matrices / vectors
    L = 2 * M + 1  # length of the (complex-valued) b vector

    sz_Tb0 = L - K  # the only effective size because of Hermitian symmetry
    sz_Tb1 = K + 1

    sz_Rc0 = L - K
    sz_Rc1 = L

    sz_coef = K + 1
    sz_bri = L

    # indices where the 4 x 4 block matrix is updated at each iteration
    # for -R(c)
    row_s1, row_e1 = sz_coef, sz_coef + sz_Tb0
    col_s1, col_e1 = sz_Tb0 + sz_Tb1, sz_Tb0 + sz_Tb1 + sz_Rc1
    # for -R(c).T
    row_s2, row_e2 = sz_coef + sz_Tb0, sz_coef + sz_Tb0 + sz_Rc1
    col_s2, col_e2 = sz_coef, sz_coef + sz_Tb0

    GtG = np.dot(G.T, G)
    D = linalg.block_diag(D1, D2)

    c0_ri_half = c_ri_half.copy()
    error_seq = np.zeros(max_iter, dtype=float)
    R_loop = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)

    # first row of mtx_loop
    mtx_loop_first_row = np.hstack((np.zeros((sz_coef, sz_coef)), Tbeta_ri.T,
                                    np.zeros((sz_coef, sz_Rc1)), c0_ri_half))
    # last row of mtx_loop
    mtx_loop_last_row = np.hstack((c0_ri_half.T, np.zeros((1, sz_Tb0 + sz_Rc1 + 1))))

    mtx_loop = np.vstack((mtx_loop_first_row,
                          np.hstack((Tbeta_ri,
                                     np.zeros((sz_Tb0, sz_Tb0)),
                                     -R_loop,
                                     np.zeros((sz_Rc0, 1))
                                     )),
                          np.hstack((np.zeros((sz_Rc1, sz_Tb1)),
                                     -R_loop.T,
                                     GtG,
                                     np.zeros((sz_Rc1, 1))
                                     )),
                          mtx_loop_last_row
                          ))

    mtx_brecon = np.zeros((sz_Rc1 + sz_Rc0, sz_Rc1 + sz_Rc0))
    mtx_brecon[:sz_Rc1, :sz_Rc1] = GtG

    for inner in range(max_iter):

        # update the mtx_loop matrix
        mtx_loop[row_s1:row_e1, col_s1:col_e1] = -R_loop
        mtx_loop[row_s2:row_e2, col_s2:col_e2] = -R_loop.T

        # matrix should be symmetric
        # mtx_loop = (mtx_loop + mtx_loop.T) / 2.
        mtx_loop += mtx_loop.T
        mtx_loop *= 0.5
        # c_ri_half = linalg.lstsq(mtx_loop, rhs)[0][:sz_coef]
        c_ri_half = linalg.solve(mtx_loop, rhs, check_finite=False)[:sz_coef]

        R_loop = Rmtx_ri_half_out_half(c_ri_half, K, D, L, D_coef, mtx_shrink)

        mtx_brecon[:sz_Rc1, sz_Rc1:] = R_loop.T
        mtx_brecon[sz_Rc1:, :sz_Rc1] = R_loop
        # mtx_brecon = (mtx_brecon + mtx_brecon.T) / 2.
        mtx_brecon += mtx_brecon.T
        mtx_brecon *= 0.5

        # b_recon_ri = linalg.lstsq(mtx_brecon, rhs_bl)[0][:sz_bri]
        b_recon_ri = linalg.solve(mtx_brecon, rhs_bl, check_finite=False)[:sz_bri]

        error_seq[inner] = linalg.norm(a_ri - np.dot(G, b_recon_ri))
        if error_seq[inner] < min_error:
            min_error = error_seq[inner]
            b_opt = np.dot(D1, b_recon_ri[:M + 1]) + 1j * np.dot(D2, b_recon_ri[M + 1:])
            c_ri = np.dot(D_coef, c_ri_half)
            c_opt = c_ri[:K + 1] + 1j * c_ri[K + 1:]  # real and imaginary parts
    return c_opt, min_error, b_opt

def make_G(p_mic_x, p_mic_y, omega_bands, sound_speed, M, signal_type='visibility'):
    """
    reconstruct point sources on the circle from the visibility measurements
    from multi-bands.

    Parameters
    ----------
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    omega_bands: 
        mid-band (ANGULAR) frequencies [radian/sec]
    sound_speed: 
        speed of sound
    signal_type: 
        The type of the signal a, possible values are 'visibility' for
        covariance matrix and 'raw' for microphone inputs

    Returns
    -------
    The list of mapping matrices from measurements to sinusoids
    """
    
    # expansion matrices to take Hermitian symmetry into account
    norm_factor = np.reshape(sound_speed / omega_bands, (1, -1), order='F')
    # normalised antenna coordinates in a matrix form
    # each column corresponds to the location in one subband,
    # i.e., the second dimension corresponds to different subbands
    p_mic_x_normalised = np.reshape(p_mic_x, (-1, 1), order='F') / norm_factor
    p_mic_y_normalised = np.reshape(p_mic_y, (-1, 1), order='F') / norm_factor

    D1, D2 = hermitian_expan(M + 1)
    # G = mtx_fri2visi_ri_multiband(M, p_mic_x_normalised, p_mic_y_normalised, D1, D2)
    G_lst = mtx_fri2signal_ri_multiband(M, p_mic_x_normalised, p_mic_y_normalised,
                                        D1, D2, aslist=True, signal=signal_type)

    return np.array(G_lst)


def make_GtG_and_inv(G_lst):

    GtG_lst = []
    GtG_inv_lst = []
    for loop in range(len(G_lst)):
        G_loop = G_lst[loop]

        GtG = np.dot(G_loop.T, G_loop)


        GtG_lst.append(GtG)
        GtG_inv_lst.append(linalg.inv(GtG))

    return np.array(GtG_lst), np.array(GtG_inv_lst)


def pt_src_recon_multiband(a, p_mic_x, p_mic_y, omega_bands, sound_speed,
                           K, M, noise_level, max_ini=50,
                           update_G=False, verbose=False, signal_type='visibility', 
                           max_iter=50, 
                           G_lst=None, GtG_lst=None, GtG_inv_lst=None, 
                           **kwargs):
    """
    reconstruct point sources on the circle from the visibility measurements
    from multi-bands.

    Parameters
    ----------
    a: 
        the measured visibilities in a matrix form, where the second dimension
        corresponds to different subbands
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    omega_bands: 
        mid-band (ANGULAR) frequencies [radian/sec]
    sound_speed: 
        speed of sound
    K: 
        number of point sources
    M: 
        the Fourier series expansion is between -M to M
    noise_level: 
        noise level in the measured visibilities
    max_ini: 
        maximum number of random initialisation used
    update_G: 
        update the linear mapping that links the uniformly sampled sinusoids to
        the visibility or not.
    verbose: 
        whether output intermediate results for debugging or not
    signal_type: 
        The type of the signal a, possible values are 'visibility' for
        covariance matrix and 'raw' for microphone inputs
    kwargs: 
        possible optional input: G_iter: number of iterations for the G updates
    """

    p_mic_x = np.squeeze(p_mic_x)
    p_mic_y = np.squeeze(p_mic_y)

    num_bands = np.array(omega_bands).size  # number of bands considered
    num_mic = p_mic_x.size

    assert M >= K
    assert num_mic * (num_mic - 1) >= 2 * M + 1
    if len(a.shape) == 2:
        assert a.shape[1] == num_bands

    a_ri = np.row_stack((a.real, a.imag))

    if update_G:
        if 'G_iter' in kwargs:
            max_loop_G = kwargs['G_iter']
        else:
            max_loop_G = 2
    else:
        max_loop_G = 1

    # initialisation
    min_error = float('inf')
    phik_opt = np.zeros(K)
    alphak_opt = np.zeros(K)

    # expansion matrices to take Hermitian symmetry into account
    norm_factor = np.reshape(sound_speed / omega_bands, (1, -1), order='F')
    # normalised antenna coordinates in a matrix form
    # each column corresponds to the location in one subband,
    # i.e., the second dimension corresponds to different subbands
    p_mic_x_normalised = np.reshape(p_mic_x, (-1, 1), order='F') / norm_factor
    p_mic_y_normalised = np.reshape(p_mic_y, (-1, 1), order='F') / norm_factor

    D1, D2 = hermitian_expan(M + 1)

    # Create G if necessary
    if G_lst is None:
        G_lst = make_G(p_mic_x, p_mic_y, omega_bands, sound_speed, M, signal_type=signal_type)
        GtG_lst, GtG_inv_lst = make_GtG_and_inv(G_lst)

    if GtG_lst is None or GtG_inv_lst is None:
        GtG_lst, GtG_inv_lst = make_GtG_and_inv(G_lst)

    for loop_G in range(max_loop_G):
        # c_recon, error_recon = \
        #     dirac_recon_ri_half_multiband_parallel(G, a_ri, K, M, max_ini)[:2]

        # tic = time.time()
        '''the original implementation'''
        # c_recon, error_recon = \
        #     dirac_recon_ri_half_multiband(G_lst, a_ri, K, M, max_ini)[:2]

        '''faster version with lu decomposition'''
        c_recon, error_recon = \
            dirac_recon_ri_half_multiband_lu(G_lst, GtG_lst, GtG_inv_lst,
                    a_ri, K, M, max_ini, max_iter=max_iter)[:2]
        # toc = time.time()
        # print(toc - tic)

        if verbose:
            print('noise level: {0:.3e}'.format(noise_level))
            print('objective function value: {0:.3e}'.format(error_recon))
        # recover Dirac locations
        uk = np.roots(np.squeeze(c_recon))
        uk /= np.abs(uk)
        phik_recon = np.mod(-np.angle(uk), 2. * np.pi)

        # use least square to reconstruct amplitudes
        if signal_type == 'visibility':
            error_loop = 0
            alphak_recon = []
            for band_count in range(num_bands):
                amp_mtx_ri_loop = build_mtx_amp_ri(
                    p_mic_x_normalised[:, band_count],
                    p_mic_y_normalised[:, band_count],
                    phik_recon
                )

                amplitude_band = \
                    sp.optimize.nnls(
                        np.dot(amp_mtx_ri_loop.T, amp_mtx_ri_loop),
                        np.dot(amp_mtx_ri_loop.T, a_ri[:, band_count])
                    )[0]

                error_loop += linalg.norm(a_ri[:, band_count] -
                                          np.dot(amp_mtx_ri_loop, amplitude_band)
                                          )

                alphak_recon.append(amplitude_band)

            alphak_recon = np.concatenate(alphak_recon)

        elif signal_type == 'raw':
            partial_build_mtx_amp = partial(build_mtx_raw_amp, phi_k=phik_recon)
            blocks = []
            for band_count in range(num_bands):
                amp_loop = partial_build_mtx_amp(
                        p_mic_x_normalised[:,band_count],
                        p_mic_y_normalised[:,band_count],
                        )
                blocks.append(amp_loop)
            amp_mtx = linalg.block_diag(*blocks)


            alphak_recon = sp.linalg.lstsq(amp_mtx, a.flatten('F'))[0]

            error_loop = linalg.norm(a.flatten('F') - np.dot(amp_mtx, alphak_recon))

        else:
            raise ValueError('signal_type must be ''raw'' or ''visibility''.')

        if error_loop < min_error:
            min_error = error_loop
            phik_opt = phik_recon
            alphak_opt = np.reshape(alphak_recon, (-1, num_bands), order='F')

        # update the linear transformation matrix
        if update_G:
            G_lst = mtx_updated_G_multiband_new(phik_opt, M, p_mic_x_normalised,
                                                p_mic_y_normalised, G_lst, num_bands)
            GtG_lst, GtG_inv_lst = make_GtG_and_inv(G_lst)

    # convert propagation vector to DOA
    phik_doa = np.mod(phik_opt - np.pi, 2. * np.pi)
    return phik_doa, alphak_opt


def pt_src_recon(a, p_mic_x, p_mic_y, omega_band, sound_speed,
                 K, M, noise_level, max_ini=50,
                 stop_cri='mse', update_G=False, verbose=False, signal_type='visibility', **kwargs):
    """
    reconstruct point sources on the circle from the visibility measurements

    Parameters
    ----------
    a: 
        the measured visibilities
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    omega_band: 
        mid-band (ANGULAR) frequency [radian/sec]
    sound_speed: 
        speed of sound
    K: 
        number of point sources
    M: 
        the Fourier series expansion is between -M to M
    noise_level: 
        noise level in the measured visibilities
    max_ini: 
        maximum number of random initialisation used
    stop_cri: 
        either 'mse' or 'max_iter'
    update_G: 
        update the linear mapping that links the uniformly sampled sinusoids to
        the visibility or not.
    verbose: 
        whether output intermediate results for debugging or not
    signal_type: 
        The type of the signal a, possible values are 'visibility' for
        covariance matrix and 'raw' for microphone inputs
    kwargs: possible optional input: G_iter: number of iterations for the G updates
    """
    assert M >= K
    num_mic = p_mic_x.size
    assert num_mic * (num_mic - 1) >= 2 * M + 1
    a_ri = np.concatenate((a.real, a.imag))
    if update_G:
        if 'G_iter' in kwargs:
            max_loop_G = kwargs['G_iter']
        else:
            max_loop_G = 2
    else:
        max_loop_G = 1

    # initialisation
    min_error = float('inf')
    phik_opt = np.zeros(K)
    alphak_opt = np.zeros(K)

    # expansion matrices to take Hermitian symmetry into account
    p_mic_x_normalised = p_mic_x / (sound_speed / omega_band)
    p_mic_y_normalised = p_mic_y / (sound_speed / omega_band)
    D1, D2 = hermitian_expan(M + 1)
    G = mtx_fri2signal_ri(M, p_mic_x_normalised, p_mic_y_normalised, D1, D2, signal=signal_type)
    for loop_G in range(max_loop_G):
        # c_recon, error_recon = dirac_recon_ri(G, a_ri, K_est, M, noise_level, max_ini, stop_cri)[:2]
        # c_recon, error_recon = dirac_recon_ri_half(G, a_ri, K, M, noise_level, max_ini, stop_cri)[:2]
        c_recon, error_recon = dirac_recon_ri_half_parallel(G, a_ri, K, M, max_ini)[:2]

        if verbose:
            print('noise level: {0:.3e}'.format(noise_level))
            print('objective function value: {0:.3e}'.format(error_recon))
        # recover Dirac locations
        uk = np.roots(np.squeeze(c_recon))
        uk /= np.abs(uk)
        phik_recon = np.mod(-np.angle(uk), 2. * np.pi)

        # use least square to reconstruct amplitudes
        amp_mtx = build_mtx_amp(phik_recon, p_mic_x_normalised, p_mic_y_normalised)
        amp_mtx_ri = np.vstack((amp_mtx.real, amp_mtx.imag))
        alphak_recon = sp.optimize.nnls(amp_mtx_ri, a_ri.squeeze())[0]
        error_loop = linalg.norm(a.flatten('F') - np.dot(amp_mtx, alphak_recon))

        if error_loop < min_error:
            min_error = error_loop
            # amp_sort_idx = np.argsort(alphak_recon)[-K:]
            # phik_opt, alphak_opt = phik_recon[amp_sort_idx], alphak_recon[amp_sort_idx]
            phik_opt, alphak_opt = phik_recon, alphak_recon

        # update the linear transformation matrix
        if update_G:
            G = mtx_updated_G(phik_recon, M, amp_mtx_ri, G)

    return phik_opt, alphak_opt


def pt_src_recon_rotate(a, p_mic_x, p_mic_y, K, M, noise_level, max_ini=50,
                        stop_cri='mse', update_G=False, num_rotation=1,
                        verbose=False, signal_type='visibility', **kwargs):
    """
    reconstruct point sources on the circle from the visibility measurements.
    Here we apply random rotations to the coordiantes.

    Parameters
    ----------
    a: 
        the measured visibilities
    p_mic_x: 
        a vector that contains microphones' x-coordinates
    p_mic_y: 
        a vector that contains microphones' y-coordinates
    K: 
        number of point sources
    M: 
        the Fourier series expansion is between -M to M
    noise_level: 
        noise level in the measured visibilities
    max_ini: 
        maximum number of random initialisation used
    stop_cri: 
        either 'mse' or 'max_iter'
    update_G: 
        update the linear mapping that links the uniformly sampled sinusoids to
        the visibility or not.
    num_rotation: 
        number of random rotations
    verbose: 
        whether output intermediate results for debugging or not
    signal_type: 
        The type of the signal a, possible values are 'visibility' for
        covariance matrix and 'raw' for microphone inputs
    kwargs: 
        possible optional input: G_iter: number of iterations for the G updates
    """
    assert M >= K
    num_mic = p_mic_x.size
    assert num_mic * (num_mic - 1) >= 2 * M + 1
    a_ri = np.concatenate((a.real, a.imag))
    if update_G:
        if 'G_iter' in kwargs:
            max_loop_G = kwargs['G_iter']
        else:
            max_loop_G = 2
    else:
        max_loop_G = 1

    # expansion matrices to take Hermitian symmetry into account
    D1, D2 = hermitian_expan(M + 1)

    # initialisation
    min_error = float('inf')
    phik_opt = np.zeros(K)
    alphak_opt = np.zeros(K)

    # random rotation angles
    rotate_angle_all = np.random.rand(num_rotation) * np.pi * 2.
    for rand_rotate in range(num_rotation):
        rotate_angle_loop = rotate_angle_all[rand_rotate]
        rotate_mtx = np.array([[np.cos(rotate_angle_loop), -np.sin(rotate_angle_loop)],
                               [np.sin(rotate_angle_loop), np.cos(rotate_angle_loop)]])

        # rotate microphone steering vector
        # (due to change of coordinate w.r.t. the random rotation)
        p_mic_xy_rotated = np.dot(rotate_mtx,
                                  np.row_stack((p_mic_x.flatten('F'),
                                                p_mic_y.flatten('F')))
                                  )
        p_mic_x_rotated = np.reshape(p_mic_xy_rotated[0, :], p_mic_x.shape, order='F')
        p_mic_y_rotated = np.reshape(p_mic_xy_rotated[1, :], p_mic_y.shape, order='F')

        # linear transformation matrix that maps uniform samples
        # of sinusoids to visibilities
        G = mtx_fri2signal_ri(M, p_mic_x_rotated, p_mic_y_rotated, D1, D2, signal=signal_type)

        for loop_G in range(max_loop_G):
            c_recon, error_recon = dirac_recon_ri_half(G, a_ri, K, M, noise_level,
                                                       max_ini, stop_cri)[:2]
            if verbose:
                print('noise level: {0:.3e}'.format(noise_level))
                print('objective function value: {0:.3e}'.format(error_recon))

            # recover Dirac locations
            uk = np.roots(np.squeeze(c_recon))
            uk /= np.abs(uk)
            phik_recon_rotated = np.mod(-np.angle(uk), 2. * np.pi)

            # use least square to reconstruct amplitudes
            amp_mtx = build_mtx_amp(phik_recon_rotated, p_mic_x_rotated, p_mic_y_rotated)
            amp_mtx_ri = np.vstack((amp_mtx.real, amp_mtx.imag))
            alphak_recon = sp.optimize.nnls(amp_mtx_ri, a_ri.squeeze())[0]
            error_loop = linalg.norm(a.flatten('F') - np.dot(amp_mtx, alphak_recon))

            if error_loop < min_error:
                min_error = error_loop

                # rotate back
                # transform to cartesian coordinate
                xk_recon_rotated, yk_recon_rotated = polar2cart(1, phik_recon_rotated)
                xy_rotate_back = linalg.solve(rotate_mtx,
                                              np.row_stack((xk_recon_rotated.flatten('F'),
                                                            yk_recon_rotated.flatten('F')))
                                              )
                xk_recon = np.reshape(xy_rotate_back[0, :], xk_recon_rotated.shape, 'F')
                yk_recon = np.reshape(xy_rotate_back[1, :], yk_recon_rotated.shape, 'F')
                # transform back to polar coordinate
                phik_recon = np.mod(np.arctan2(yk_recon, xk_recon), 2. * np.pi)

                phik_opt, alphak_opt = phik_recon, alphak_recon

            # update the linear transformation matrix
            if update_G:
                G = mtx_updated_G(phik_recon_rotated, M, amp_mtx_ri, G)

    return phik_opt, alphak_opt



import numpy as np
from pyroomacoustics import autocorr
from unittest import TestCase
import time

N = 256
n_iter = 100
x = np.random.randn(N)
tol = 1e-12


def consistent_results(p, biased=True):

    r_time = autocorr(x, p, method='time', biased=biased)
    r_fft = autocorr(x, p, method='fft', biased=biased)
    r_np = autocorr(x, p, method='numpy', biased=biased)
    r_pra = autocorr(x, p, method='pra', biased=biased)

    err_fft = np.linalg.norm(r_time - r_fft)
    err_np = np.linalg.norm(r_time - r_np)
    err_pra = np.linalg.norm(r_time - r_pra)
    return err_fft, err_np, err_pra


class TestAutoCorr(TestCase):

    def test_consistent_low_p_biased(self):
        res = consistent_results(p=3)
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        self.assertTrue(res[2] < tol)

    def test_consistent_high_p_biased(self):
        res = consistent_results(p=35)
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        self.assertTrue(res[2] < tol)

    def test_consistent_low_p_unbiased(self):
        res = consistent_results(p=3, biased=False)
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        self.assertTrue(res[2] < tol)

    def test_consistent_high_p_unbiased(self):
        res = consistent_results(p=35, biased=False)
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        self.assertTrue(res[2] < tol)


def timing_and_error(p):
    print()
    print("-" * 40)
    print("p = {}".format(p))
    print("-" * 40)

    print("TIMING:")

    start_time = time.time()
    for n in range(n_iter):
        autocorr(x, p, method="time")
    proc_time = (time.time() - start_time) / n_iter * 1e6
    print("time   : {} us".format(proc_time))

    start_time = time.time()
    for n in range(n_iter):
        autocorr(x, p, method="fft")
    proc_time = (time.time() - start_time) / n_iter * 1e6
    print("fft    : {} us".format(proc_time))

    start_time = time.time()
    for n in range(n_iter):
        autocorr(x, p, method="numpy")
    proc_time = (time.time() - start_time) / n_iter * 1e6
    print("numpy  : {} us".format(proc_time))

    start_time = time.time()
    for n in range(n_iter):
        autocorr(x, p, method="pra")
    proc_time = (time.time() - start_time) / n_iter * 1e6
    print("pra    : {} us".format(proc_time))

    print()
    print("BIASED")

    res = consistent_results(p, biased=True)
    print("fft   error wrt time : {}".format(res[0]))
    print("numpy error wrt time : {}".format(res[1]))
    print("pra   error wrt time : {}".format(res[2]))

    print()
    print("UNBIASED")

    res = consistent_results(p, biased=False)
    print("fft   error wrt time : {}".format(res[0]))
    print("numpy error wrt time : {}".format(res[1]))
    print("pra   error wrt time : {}".format(res[2]))


if __name__ == '__main__':
    timing_and_error(p=3)
    timing_and_error(p=35)







import unittest
import numpy as np
import pyroomacoustics as pra

room_dim = [15, 14, 16]
absorption = 0.2
source_position = [2.0, 3.1, 2.0]
mic_position = [2.0, 1.5, 2.0]
fs = 16000
max_order = 5

# scenario A
def get_room_constructor_args():
    '''
    When provided with sources and microphones, the constructor
    should try to compute the RIR immediately
    '''
    source = pra.SoundSource(position=source_position)
    mics = pra.MicrophoneArray(np.array([mic_position]).T, fs)
    shoebox = pra.ShoeBox(
            room_dim,
            absorption=absorption,
            fs=fs,
            max_order=max_order,
            sources=[source],
            mics=mics,
            )

    shoebox.image_source_model()
    shoebox.compute_rir()
    return shoebox

#scenario B
def get_room_add_method():
    shoebox = pra.ShoeBox(room_dim, absorption=absorption, fs=fs, max_order=max_order)
    shoebox.add_source(source_position)
    mics = pra.MicrophoneArray(np.array([mic_position]).T, fs)
    shoebox.add_microphone_array(mics)

    shoebox.image_source_model()
    shoebox.compute_rir()
    return shoebox

class RoomConstructorSources(unittest.TestCase):

    def test_room_constructor(self):

        room_1 = get_room_constructor_args()
        self.assertTrue(isinstance(room_1.sources[0], pra.SoundSource))

    def test_room_add_method(self):
        room_2 = get_room_add_method()
        self.assertTrue(isinstance(room_2.sources[0], pra.SoundSource))

    def test_rir_equal(self):
        room_1 = get_room_constructor_args()
        room_2 = get_room_add_method()
        self.assertTrue(np.allclose(room_1.rir[0][0], room_2.rir[0][0]))


if __name__ == '__main__':

    unittest.main()



from __future__ import division, print_function
from unittest import TestCase
import numpy as np
import pyroomacoustics as pra

tol = -80  # dB
nfft = 128
D = 7
x = np.random.randn(nfft, D).astype('float32')
X_numpy = np.fft.rfft(x, axis=0).astype('complex64')
analysis_window = pra.hann(nfft)
synthesis_window = pra.hann(nfft)

eps = np.finfo(x.dtype).eps

try:
    import pyfftw
    pyfftw_available = True
except ImportError:
    pyfftw_available = False

def no_window(nfft, D, transform, axis=0):

    if D == 1:
        x_local = x[:,0]
        X_local = X_numpy[:,0]
    else:
        if axis == 0:
            x_local = x
            X_local = X_numpy
        else:
            x_local = x.T
            X_local = X_numpy.T

    # make object
    dft = pra.transform.DFT(nfft, D, transform=transform, axis=axis)

    # forward
    X = dft.analysis(x_local)
    err_fwd = pra.dB(np.linalg.norm(X_local-X) + eps)

    # backward
    x_r = dft.synthesis()
    err_bwd = pra.dB(np.linalg.norm(x_local-x_r) + eps)

    return err_fwd, err_bwd


def window(nfft, D, analysis_window, synthesis_window, axis=0):

    if D == 1:
        x_local = x[:,0]
        X_local = X_numpy[:,0]
    else:
        if axis == 0:
            x_local = x
            X_local = X_numpy
        else:
            x_local = x.T
            X_local = X_numpy.T

    # make object
    dft = pra.transform.DFT(nfft, D, axis=axis, analysis_window=analysis_window,
                            synthesis_window=synthesis_window)

    try:
        # forward
        X = dft.analysis(x_local)
        # backward
        x_r = dft.synthesis()
        return True
    except:
        return False


class TestDFT(TestCase):

    def test_no_window_mono(self):
        res = no_window(nfft, D=1, transform='numpy')
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        if pyfftw_available:
            res = no_window(nfft, D=1, transform='fftw')
            self.assertTrue(res[0] < tol)
            self.assertTrue(res[1] < tol)
        res = no_window(nfft, D=1, transform='mkl')
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)

    def test_no_window_multichannel_axis0(self):
        res = no_window(nfft, D=D, transform='numpy')
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        if pyfftw_available:
            res = no_window(nfft, D=D, transform='fftw')
            self.assertTrue(res[0] < tol)
            self.assertTrue(res[1] < tol)
        res = no_window(nfft, D=D, transform='mkl')
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)

    def test_no_window_multichannel_axis1(self):
        res = no_window(nfft, D=D, transform='numpy', axis=1)
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)
        if pyfftw_available:
            res = no_window(nfft, D=D, transform='fftw', axis=1)
            self.assertTrue(res[0] < tol)
            self.assertTrue(res[1] < tol)
        res = no_window(nfft, D=D, transform='mkl', axis=1)
        self.assertTrue(res[0] < tol)
        self.assertTrue(res[1] < tol)

    def test_window(self):
        res = window(nfft, 1, analysis_window, synthesis_window)
        self.assertTrue(res)
        res = window(nfft, D, analysis_window, synthesis_window)
        self.assertTrue(res)
        res = window(nfft, D, analysis_window, synthesis_window, axis=1)
        self.assertTrue(res)


if __name__ == "__main__":

    print()
    print("1D")
    res = no_window(nfft, D=1, transform='numpy')
    print("numpy :", res)
    if pyfftw_available:
        res = no_window(nfft, D=1, transform='fftw')
        print("fftw :", res)
    res = no_window(nfft, D=1, transform='mkl')
    print("mkl :", res)

    print()
    print("2D, axis=0")
    res = no_window(nfft, D=D, transform='numpy')
    print("numpy :", res)
    if pyfftw_available:
        res = no_window(nfft, D=D, transform='fftw')
        print("fftw :", res)
    res = no_window(nfft, D=D, transform='mkl')
    print("mkl :", res)

    print()
    print("2D, axis=1")
    axis=1
    res = no_window(nfft, D=D, transform='numpy', axis=axis)
    print("numpy :", res)
    if pyfftw_available:
        res = no_window(nfft, D=D, transform='fftw', axis=axis)
        print("fftw :", res)
    res = no_window(nfft, D=D, transform='mkl', axis=axis)
    print("mkl :", res)

    print()
    print("Testing no error with windows...")
    res = window(nfft, 1, analysis_window, synthesis_window)
    print(res)
    res = window(nfft, D, analysis_window, synthesis_window)
    print(res)
    res = window(nfft, D, analysis_window, synthesis_window, axis=1)
    print(res)





# Test of ray tracing energy
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import unittest
import numpy as np
import pyroomacoustics as pra


class SimpleHistogram(list):
    """ A Histogram class based on list """

    def __init__(self, bin_size):
        self.bin_size = bin_size
        list.__init__([])

    def add(self, delay, val):
        """
        Adds val to the bin corresponding to delay.
        Pad the histogram to correct size if too short.
        """
        pos = int(delay / self.bin_size)

        if pos >= len(self):
            while len(self) < pos:
                self.append(0)
            self.append(val)
        else:
            self[pos] += val


class TestRayEnergy(unittest.TestCase):
    def test_square_room(self):

        """
        This is a cubic room of 2x2x2 meters. The source is placed at [0.5,0.5, 1]
        and the receiver at [1.5, 1.5, 1]. A ray is launched towards [1, 0, 1] so that
        the first receiver hit is after travel distance of 2*sqrt(2) and each subsequent
        hit travels a further 4*sqrt(2) until the threshold energy is reached.
        """

        energy_absorption = 0.07
        round_trip = 4 * np.sqrt(2)
        energy_thresh = 1e-7
        detector_radius = 0.15
        hist_bin_size = 0.004  # resolution of histogram [s]

        histogram_gt = SimpleHistogram(hist_bin_size * pra.constants.get("c"))

        # Create the groundtruth list of energy and travel time
        initial_energy = 2.0  # defined in libroom.Room.get_rir_entries
        transmitted = 1.0 * (1.0 - energy_absorption) ** 2 * initial_energy
        distance = round_trip / 2.0

        while transmitted / distance > energy_thresh:
            r_sq = distance ** 2
            p_hit = 1.0 - np.sqrt(1.0 - detector_radius ** 2 / r_sq)
            histogram_gt.add(distance, transmitted / (r_sq * p_hit))
            transmitted *= (1.0 - energy_absorption) ** 4  # 4 wall hits
            distance += round_trip

        print("Creating the python room (polyhedral)")
        walls_corners = [
            np.array([[0, 2, 2, 0], [0, 0, 0, 0], [0, 0, 2, 2]]),  # left
            np.array([[0, 0, 2, 2], [2, 2, 2, 2], [0, 2, 2, 0]]),  # right
            np.array([[0, 0, 0, 0], [0, 2, 2, 0], [0, 0, 2, 2]]),  # front`
            np.array([[2, 2, 2, 2], [0, 0, 2, 2], [0, 2, 2, 0]]),  # back
            np.array([[0, 2, 2, 0], [0, 0, 2, 2], [0, 0, 0, 0],]),  # floor
            np.array([[0, 0, 2, 2], [0, 2, 2, 0], [2, 2, 2, 2],]),  # ceiling
        ]
        walls = [pra.wall_factory(c, [energy_absorption], [0.0]) for c in walls_corners]
        room_poly = pra.Room(walls, fs=16000)
        # room = pra.Room(walls, fs=16000)
        room_poly.add_source([0.5, 0.5, 1])
        room_poly.add_microphone_array(
            pra.MicrophoneArray(np.c_[[1.5, 1.5, 1.0]], room_poly.fs)
        )

        room_poly.room_engine.set_params(
            room_poly.c,
            0,
            energy_thresh,  # energy threshold for rays
            5.0,  # time threshold for rays
            detector_radius,  # detector radius
            hist_bin_size,  # resolution of histogram [s]
            False,  # is it hybrid model ?
        )

        print("Running ray tracing (polyhedral)")
        room_poly.room_engine.ray_tracing(
            np.c_[[-np.pi / 4.0, np.pi / 2.0]],
            room_poly.sources[0].position,  # source loc
        )

        print("Creating the python room (shoebox)")
        room_cube = pra.ShoeBox(
            [2, 2, 2], fs=16000, materials=pra.Material(energy_absorption)
        )
        # room = pra.Room(walls, fs=16000)
        room_cube.add_source([0.5, 0.5, 1])
        room_cube.add_microphone_array(
            pra.MicrophoneArray(np.c_[[1.5, 1.5, 1.0]], room_poly.fs)
        )

        room_cube.room_engine.set_params(
            room_poly.c,
            0,
            energy_thresh,  # energy threshold for rays
            5.0,  # time threshold for rays
            detector_radius,  # detector radius
            hist_bin_size,  # resolution of histogram [s]
            False,  # is it hybrid model ?
        )

        print("Running ray tracing (shoebox)")
        room_cube.room_engine.ray_tracing(
            np.c_[[-np.pi / 4.0, np.pi / 2.0]],
            room_cube.sources[0].position,  # source loc
        )

        h_poly = room_poly.room_engine.microphones[0].histograms[0].get_hist()
        h_cube = room_cube.room_engine.microphones[0].histograms[0].get_hist()
        histogram_rt_poly = np.array(h_poly[0])[: len(histogram_gt)]
        histogram_rt_cube = np.array(h_cube[0])[: len(histogram_gt)]

        self.assertTrue(np.allclose(histogram_rt_poly, histogram_gt))
        self.assertTrue(np.allclose(histogram_rt_cube, histogram_gt))


if __name__ == "__main__":
    unittest.main()



'''
A few test signals like sweeps and stuff.
'''
from __future__ import division, print_function
import numpy as np

def window(signal, n_win):
    ''' window the signal at beginning and end with window of size n_win/2 '''

    win = np.hanning(2*n_win)

    sig_copy = signal.copy()

    sig_copy[:n_win] *= win[:n_win]
    sig_copy[-n_win:] *= win[-n_win:]

    return sig_copy

def exponential_sweep(T, fs, f_lo=0., f_hi=None, fade=None, ascending=False):
    '''
    Exponential sine sweep

    Parameters
    ----------
    T: float
        length in seconds
    fs: 
        sampling frequency
    f_lo: float
        lowest frequency in fraction of fs (default 0)
    f_hi: float
        lowest frequency in fraction of fs (default 1)
    fade: float, optional
        length of fade in and out in seconds (default 0)
    ascending: bool, optional
    '''

    if f_hi is None:
        f_hi = fs / 2
    elif f_hi < 0.:
        f_hi = fs / 2 + f_hi
    elif f_hi > fs / 2:
        f_hi = fs / 2

    if f_lo < 1.:
        f_lo = 1.

    if f_lo > f_hi:
        raise ValueError('Error: need 0. <= f_lo < f_hi <= fs/2')

    Ts = 1./fs   # Sampling period in [s]
    N = np.floor(T/Ts)  # number of samples
    n  = np.arange(0, N, dtype='float64')  # Sample index

    om1 = 2 * np.pi * f_lo
    om2 = 2 * np.pi * f_hi

    sweep = np.sin(om1*N*Ts / np.log(om2/om1) * (np.exp(n/N*np.log(om2/om1)) - 1))

    if not ascending:
        sweep = sweep[::-1]

    if fade is not None and fade > 0.:
        sweep = window(sweep, int(fs * fade))

    return sweep

def linear_sweep(T, fs, f_lo=0., f_hi=None, fade=None, ascending=False):
    '''
    Linear sine sweep

    Parameters
    ----------
    T: float
        length in seconds
    fs: 
        sampling frequency
    f_lo: float
        lowest frequency in fraction of fs (default 0)
    f_hi: float
        lowest frequency in fraction of fs (default 1)
    fade: float, optional
        length of fade in and out in seconds (default 0)
    ascending: bool, optional
    '''

    if f_hi is None:
        f_hi = fs / 2
    elif f_hi < 0.:
        f_hi = fs / 2 + f_hi
    elif f_hi > fs / 2:
        f_hi = fs / 2

    if f_lo < 0.:
        f_lo = 0.

    if f_lo > f_hi:
        raise ValueError('Error: need 0. <= f_lo < f_hi <= fs/2')

    Ts = 1./fs   # Sampling period in [s]

    N = np.floor(T/Ts)  # number of samples
    n  = np.arange(0, N, dtype='float64')  # Sample index

    t = n * Ts  # time vector
    rate = (f_hi - f_lo) / T  # rate of change of frequency

    sweep = np.sin(2 * np.pi * (f_lo + 0.5 * rate * t) * t)

    if not ascending:
        sweep = sweep[::-1]

    if fade is not None and fade > 0.:
        sweep = window(sweep, int(fs * fade))

    return sweep




from __future__ import division, print_function
from unittest import TestCase
import numpy as np
import pyroomacoustics as pra
from pyroomacoustics.transform import stft


# test parameters
tol = -100  # dB
np.random.seed(0)
D = 3
block_size = 512

# test signal (noise)
x = np.random.randn(block_size * 20, D).astype(np.float32)


def no_overlap(D):

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    hop = block_size

    # analysis
    X = stft.analysis(x_local, L=block_size, hop=hop)

    # synthesis
    x_r = stft.synthesis(X, L=block_size, hop=hop)

    return pra.dB(np.max(np.abs(x_local - x_r)))


def half_overlap(D):

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    hop = block_size // 2

    # analysis
    analysis_win = pra.hann(block_size)
    X = stft.analysis(x_local, L=block_size, hop=hop, win=analysis_win)

    # synthesis
    x_r = stft.synthesis(X, L=block_size, hop=hop)

    return pra.dB(
        np.max(np.abs(x_local[: -block_size + hop,] - x_r[block_size - hop :,]))
    )


def append_one_sample(D):
    hop = block_size // 2
    n_samples = x.shape[0]
    n_frames = n_samples // hop
    x_local = x[: n_frames * hop - 1, :]

    if D == 1:
        x_local = x_local[:, 0]
    else:
        x_local = x_local[:, :D]

    # analysis
    analysis_win = pra.hann(block_size)
    X = stft.analysis(x_local, L=block_size, hop=hop, win=analysis_win)

    # synthesis
    x_r = stft.synthesis(X, L=block_size, hop=hop)

    return pra.dB(
        np.max(np.abs(x_local[: -block_size + hop,] - x_r[block_size - hop : -1,]))
    )


def hop_one_sample(D):

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    hop = 1

    # analysis
    analysis_win = pra.hann(block_size)
    X = stft.analysis(x_local, L=block_size, hop=hop, win=analysis_win)

    # synthesis
    synthesis_win = pra.transform.stft.compute_synthesis_window(analysis_win, hop)
    x_r = stft.synthesis(X, L=block_size, hop=hop, win=synthesis_win)

    return pra.dB(
        np.max(np.abs(x_local[: -block_size + hop,] - x_r[block_size - hop :,]))
    )


class TestSTFTOneShot(TestCase):
    def test_no_overlap(self):
        self.assertTrue(no_overlap(1) < tol)
        self.assertTrue(no_overlap(D) < tol)

    def test_half_overlap(self):
        self.assertTrue(half_overlap(1) < tol)
        self.assertTrue(half_overlap(D) < tol)

    def test_append_one_sample(self):
        self.assertTrue(append_one_sample(1) < tol)
        self.assertTrue(append_one_sample(D) < tol)

    def test_hop_one_sample(self):
        self.assertTrue(hop_one_sample(1) < tol)
        self.assertTrue(hop_one_sample(D) < tol)


if __name__ == "__main__":

    print()
    print("TEST INFO")
    print("-------------------------------------------------------------")
    print("Max error in dB for randomly generated signal of %d samples." % len(x))
    print("Multichannel corresponds to %d channels." % D)
    print("-------------------------------------------------------------")
    print()

    err = no_overlap(1)
    print("No overlap, mono                : %d dB" % err)
    err = no_overlap(D)
    print("No overlap, multichannel        : %d dB" % err)

    err = half_overlap(1)
    print("No overlap, mono                : %d dB" % err)
    err = half_overlap(D)
    print("Half overlap, multichannel      : %d dB" % err)

    # check squeeze done properly
    err = append_one_sample(1)
    print("Append one zero, mono           : %d dB" % err)
    err = append_one_sample(D)
    print("Append one zero, multichannel   : %d dB" % err)

    err = hop_one_sample(1)
    print("Hop one sample, mono            : %d dB" % err)
    err = hop_one_sample(D)
    print("Hop one sample, multichannel    : %d dB" % err)



'''
Direction of Arrival Finding
============================

This sub-package provides implementations of popular direction of arrival findings algorithms.

MUSIC
    | Multiple Signal Classification [1]_ 
    | :py:obj:`pyroomacoustics.doa.music` 
SRP-PHAT 
    | Steered Response Power -- Phase Transform [2]_ 
    | :py:obj:`pyroomacoustics.doa.srp`
CSSM 
    | Coherent Signal Subspace Method [3]_
    | :py:obj:`pyroomacoustics.doa.cssm`
WAVES 
    | Weighted Average of Signal Subspaces [4]_ 
    | :py:obj:`pyroomacoustics.doa.waves`
TOPS 
    | Test of Orthogonality of Projected Subspaces [5]_
    | :py:obj:`pyroomacoustics.doa.tops`
FRIDA 
    | Finite Rate of Innovation Direction of Arrival [6]_ 
    | :py:obj:`pyroomacoustics.doa.frida`

All these classes derive from the abstract base class
:py:obj:`pyroomacoustics.doa.doa.DOA` that offers generic methods for finding
and visualizing the locations of acoustic sources.

The constructor can be called once to build the DOA finding object. Then, the
method :py:obj:`pyroomacoustics.doa.doa.DOA.locate_sources` performs DOA
finding based on time-frequency passed to it as an argument. Extra arguments
can be supplied to indicate which frequency bands should be used for
localization.

How to use the DOA module
-------------------------

Here ``R`` is a 2xQ ndarray that contains the locations of the Q microphones
in the columns, ``fs`` is the sampling frequency of the input signal, and
``nfft`` the length of the FFT used.

The STFT snapshots are passed to the localization methods in the X ndarray of
shape ``Q x (nfft // 2 + 1) x n_snapshots``, where ``n_snapshots`` is the
number of STFT frames to use for the localization. The option ``freq_bins``
can be provided to specify which frequency bins to use for the localization.

    >>> doa = pyroomacoustics.doa.MUSIC(R, fs, nfft)
    >>> doa.locate_sources(X, freq_bins=np.arange(20, 40))

Other Available Subpackages
---------------------------

:py:obj:`pyroomacoustics.doa.grid`
    this provides abstractions for computing functions on regular or irregular
    grids defined on circles and spheres with peak finding methods

:py:obj:`pyroomacoustics.doa.plotters`
    a few methods to plot functions and points on circles or spheres

:py:obj:`pyroomacoustics.doa.detect_peaks`
    1D peak detection routine from Marcos Duarte

:py:obj:`pyroomacoustics.doa.tools_frid_doa_plane`
    routines implementing FRIDA algorithm

Utilities
---------

:py:obj:`pyroomacoustics.doa.algorithms`
    a dictionary containing all the DOA object subclasses availables indexed by
    keys ``['MUSIC', 'SRP', 'CSSM', 'WAVES', 'TOPS', 'FRIDA']``

References
----------

.. [1] R. Schmidt, *Multiple emitter location and signal parameter estimation*, 
    IEEE Trans. Antennas Propag., Vol. 34, Num. 3, pp 276--280, 1986

.. [2] J. H. DiBiase, *A high-accuracy, low-latency technique for talker localization 
    in reverberant environments using microphone arrays*, PHD Thesis, Brown University, 2000

.. [3] H. Wang, M. Kaveh, *Coherent signal-subspace processing for the detection and 
    estimation of angles of arrival of multiple wide-band sources*, IEEE Trans. Acoust., 
    Speech, Signal Process., Vol. 33, Num. 4, pp 823--831, 1985

.. [4] E. D. di Claudio, R. Parisi, *WAVES: Weighted average of signal subspaces for 
    robust wideband direction finding*, IEEE Trans. Signal Process., Vol. 49, Num. 10, 
    2179--2191, 2001

.. [5] Y. Yeo-Sun, L. M. Kaplan, J. H. McClellan, *TOPS: New DOA estimator for wideband 
    signals*, IEEE Trans. Signal Process., Vol. 54, Num 6., pp 1977--1989, 2006

.. [6] H. Pan, R. Scheibler, E. Bezzam, I. Dokmanic, and M. Vetterli, *FRIDA:
    FRI-based DOA estimation for arbitrary array layouts*, Proc. ICASSP,
    pp 3186-3190, 2017

'''

from .doa import *
from .srp import *
from .music import *
from .cssm import *
from .waves import *
from .tops import *
from .frida import *
from .grid import *
from .utils import *

# Create this dictionary as a shortcut to different algorithms
algorithms = {
        'SRP' : SRP,
        'MUSIC' : MUSIC,
        'CSSM' : CSSM,
        'WAVES' : WAVES,
        'TOPS' : TOPS,
        'FRIDA' : FRIDA,
        }




# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

import unittest
import numpy as np

import pyroomacoustics as pra

fs = 8000
t0 = 0.
absorption = 0.90
max_order_sim = 2
sigma2_n = 5e-7

corners = np.array([[0,0,4,4],[0,4,4,0]])
room = pra.Room.from_corners(
    corners,
    absorption=absorption,
    fs=fs,
    t0=t0,
    max_order=max_order_sim,
    sigma2_awgn=sigma2_n)
    
    
room.add_source([2, 2], None, 0)

# place 3 microphones in the room
mics = pra.MicrophoneArray(np.array([[2, 5, 3, 1, 2,],
                                     [2, 5, 3, 1, 1,]]), fs)
room.add_microphone_array(mics)

# run the image source model
room.image_source_model()

# we order the sources lexicographically
ordering = np.lexsort(room.sources[0].images)
images_found = room.sources[0].images[:,ordering]
visibilities = np.array(room.visibility[0][:,ordering], dtype=bool)

# This should be all visible sources, from the 3 mic locations
sources = np.array([[  2.,  -2.,  -2.,   2.,   6.,   6.,  -6.,  -2.,   2.,   6.,  10., -2.,  -2.,   2.,   6.,   6.,   2.],
                    [ -6.,  -2.,  -2.,  -2.,  -2.,  -2.,   2.,   2.,   2.,   2.,   2.,  6.,   6.,   6.,   6.,   6.,  10.]])

# These are the visibilities for each individual microphone
visible_middle = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=bool)
visible_outside = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool)
visible_upperRight = np.array([1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1], dtype=bool)
visible_lowerLeft = np.array([1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1], dtype=bool)
visible_lowMiddle = np.array([1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1], dtype=bool)


class TestVisibilityShoeBox2D(unittest.TestCase):

    def test_sources(self):
        self.assertTrue(np.allclose(images_found, sources))

    def test_visibility_middle(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[0,:]], sources[:,visible_middle]))
        
    def test_visibility_outside(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[1,:]], sources[:,visible_outside]))

    def test_visibility_upperRight(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[2,:]], sources[:,visible_upperRight]))
        
    def test_visibility_lowerLeft(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[3,:]], sources[:,visible_lowerLeft]))
        
    def test_visibility_lowMiddle(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[4,:]], sources[:,visible_lowMiddle]))

if __name__ == '__main__':
    unittest.main()



'''
Base class for some data corpus and the samples it contains.
'''

from itertools import islice

def _take(n, iterable):
    "Return first n items of the iterable as a list"
    return list(islice(iterable, n))

class Meta(object):
    '''
    A simple class that will take a dictionary as input
    and put the values in attributes named after the keys.
    We use it to store metadata for the samples

    The parameters can be any set of keyword arguments.
    They will all be transformed into attribute of the object.
    '''
    def __init__(self, **attr):
        for key, val in attr.items():
            self.__setattr__(key, val)

    def match(self, **kwargs):
        '''
        The key/value pairs given by the keyword arguments are compared
        to the attribute/value pairs of the object. If the values all
        match, True is returned. Otherwise False is returned. If a keyword
        argument has no attribute counterpart, an error is raised. Attributes
        that do not have a keyword argument counterpart are ignored.

        There are three ways to match an attribute with keyword=value:
        1. ``value == attribute``
        2. ``value`` is a list and ``attribute in value == True``
        3. ``value`` is a callable (a function) and ``value(attribute) == True``
        '''
        for key, val in kwargs.items():
            attr = self.__getattribute__(key)
            if callable(val) and val(attr):
                continue
            if isinstance(val, list) and attr in val:
                continue
            if attr == val:
                continue
            return False
            '''
            if (not (callable(val) val(attr))
                    and not (isinstance(val, list) and attr in val)
                    and attr != val):
                return False
            if attr != val and not (isinstance(val, list) and attr in val):
                return False
            '''
        return True

    def as_dict(self):
        ''' Returns all the attribute/value pairs of the object as a dictionary '''
        return self.__dict__.copy()

    def __str__(self):
        r = 'Metadata:\n'
        for attr, val in self.__dict__.items():
            r += '    {} : {}\n'.format(attr, val)
        return r[:-1]  # remove the trailing '\n'

    def __repr__(self):
        return self.__dict__.__repr__()


class Sample(object):
    '''
    The base class for a dataset sample. The idea is that different
    corpus will have different attributes for the samples. They
    should at least have a data attribute.

    Attributes
    ----------
    data: array_like
        The actual data
    meta: pyroomacoustics.datasets.Meta
        An object containing the sample metadata. They can be accessed using the
        dot operator
    '''

    def __init__(self, data, **kwargs):
        ''' Dummy init method '''
        self.data = data
        self.meta = Meta(**kwargs)

    def __str__(self):
        r = 'Data : ' + self.data.__str__() + '\n'
        r += self.meta.__str__()
        return r


class AudioSample(Sample):
    '''
    We add some methods specific to display and listen to audio samples.
    The sampling frequency of the samples is an extra parameter.

    For multichannel audio, we assume the same format used by 
    ```scipy.io.wavfile <https://docs.scipy.org/doc/scipy-0.14.0/reference/io.html#module-scipy.io.wavfile>`_``,
    that is ``data`` is then a 2D array with each column being a channel.

    Attributes
    ----------
    data: array_like
        The actual data
    fs: int
        The sampling frequency of the input signal
    meta: pyroomacoustics.datasets.Meta
        An object containing the sample metadata. They can be accessed using the
        dot operator
    '''
    def __init__(self, data, fs, **kwargs):
        Sample.__init__(self, data, **kwargs)
        self.fs = fs

    def play(self, **kwargs):
        '''
        Play the sound sample. This function uses the 
        `sounddevice <https://python-sounddevice.readthedocs.io>`_ package for playback.

        It takes the same keyword arguments as 
        `sounddevice.play <https://python-sounddevice.readthedocs.io/en/0.3.10/#sounddevice.play>`_.
        '''
        try:
            import sounddevice as sd
        except ImportError as e:
            print('Warning: sounddevice package is required to play audiofiles.')
            return

        sd.play(self.data, samplerate=self.fs, **kwargs)

    def plot(self, NFFT=512, noverlap=384, **kwargs):
        '''
        Plot the spectrogram of the audio sample. 

        It takes the same keyword arguments as
        `matplotlib.pyplot.specgram <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.specgram.html>`_.
        '''

        import numpy as np
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print('Warning: matplotlib is required for plotting')
            return

        # Handle single channel case
        if self.data.ndim == 1:
            data = self.data[:,None]

        nchannels = data.shape[1]

        # Try to have a square looking plot
        pcols = int(np.ceil(np.sqrt(nchannels)))
        prows = int(np.ceil(nchannels / pcols))

        for c in range(nchannels):
            plt.specgram(data[:,c], NFFT=NFFT, Fs=self.fs, noverlap=noverlap, **kwargs)
            plt.xlabel('Time [s]')
            plt.ylabel('Frequency [Hz]')
            plt.title('Channel {}'.format(c+1))


class Dataset(object):
    '''
    The base class for a data corpus. It has basically a list of
    samples and a filter function

    Attributes
    ----------
    samples: list
        A list of all the Samples in the dataset
    info: dict
        This dictionary keeps track of all the fields
        in the metadata. The keys of the dictionary are
        the metadata field names. The values are again dictionaries,
        but with the keys being the possible values taken by the 
        metadata and the associated value, the number of samples
        with this value in the corpus.
    '''
    def __init__(self):
        self.samples = []
        self.info = {}

    def add_sample(self, sample):
        ''' 
        Add a sample to the Dataset and keep track of the metadata.
        '''
        # keep track of the metadata going in the corpus
        for key, val in sample.meta.__dict__.items():
            if key not in self.info:
                self.info[key] = {}

            if val not in self.info[key]:
                self.info[key][val] = 1
            else:
                self.info[key][val] += 1

        # add the sample to the list
        self.samples.append(sample)

    def add_sample_matching(self, sample, **kwargs):
        '''
        The sample is added to the corpus only if all the keyword arguments
        match the metadata of the sample.  The match is operated by
        ``pyroomacoustics.datasets.Meta.match``.
        '''
        # check if the keyword arguments are matching
        if sample.meta.match(**kwargs):
            self.add_sample(sample)

    def filter(self, **kwargs):
        '''
        Filter the corpus and selects samples that match the criterias provided
        The arguments to the keyword can be 1) a string, 2) a list of strings, 3)
        a function. There is a match if one of the following is True.

        1. ``value == attribute``
        2. ``value`` is a list and ``attribute in value == True``
        3. ``value`` is a callable (a function) and ``value(attribute) == True``
        '''

        new_corpus = Dataset()

        for s in self.samples:
            new_corpus.add_sample_matching(s, **kwargs)

        return new_corpus
        
    def __getitem__(self, r):
        return self.samples[r]

    def __len__(self):
        return len(self.samples)

    def head(self, n=5):
        ''' Print n samples from the dataset '''
        print('The first', n, '/', len(self.samples), 'samples:')
        for sample in self.samples[:n]:
            print(sample)

    def __str__(self):
        r = 'The dataset contains {} samples.\n'.format(len(self))
        r += 'Metadata attributes are:\n'
        for field, values in self.info.items():

            r += '  {} ({}) :\n'.format(field, len(values))

            # for attributes with lots of values, we just print a few
            if len(values) > 6:
                short_list = _take(6, values.items())
                for value, number in short_list[:3]:
                    r += '      * {} occurs {} times\n'.format(value, number)
                r += '      ...\n'
                for value, number in short_list[3:]:
                    r += '      * {} occurs {} times\n'.format(value, number)

            else:
                for value, number in values.items():
                    r += '      * {} occurs {} times\n'.format(value, number)

        return r[:-1]  # remove trailing '\n'








'''
Point Clouds
============
Contains PointCloud class.

Given a number of points and their relative distances, this class aims at
reconstructing their relative coordinates.
'''

from __future__ import division, print_function

# Provided by LCAV
import numpy as np
from scipy import linalg as la


class PointCloud:

    def __init__(self, m=1, dim=3, diameter=0., X=None, labels=None, EDM=None):
        '''
        Parameters
        ----------

        m : int, optional
            Number of markers
        diameter : float, optional
            Diameter of the marker points (added to distance measurements)
        dim : int, optional
            Dimension of ambient space (default 3)
        X : ndarray, optional
            Array of column vectors with locations of markers
        '''

        # set the marker diameter
        self.diameter = diameter

        if EDM is not None:
            self.dim = dim
            self.fromEDM(EDM, labels=labels)

        elif X is not None:
            self.m = X.shape[1]
            self.dim = X.shape[0]
            self.X = X

            if labels is None:
                self.labels = [str(i) for i in range(self.m)]

        else:
            self.m = m
            self.dim = dim
            self.X = np.zeros((self.dim, self.m))

        # Now set the labels
        if labels is not None:
            if len(labels) == self.m:
                self.labels = labels
            else:
                raise ValueError('There needs to be one label per marker point')
        else:
            self.labels = [str(i) for i in range(self.m)]


    def __getitem__(self,ref):

        if isinstance(ref, (str, unicode)):
            if self.labels is None:
                raise ValueError('Labels not set for this marker set')
            index = self.labels.index(ref)
        elif isinstance(ref, int) or isinstance(ref, slice):
            index = ref
        elif isinstance(ref, list):
            index = [self.labels.index(s) if isinstance(s, (str, unicode)) else s for s in ref]
        else:
            index = int(ref)

        return self.X[:,index]

    def copy(self):
        ''' Return a deep copy of this marker set object '''

        new_marker = PointCloud(X=self.X.copy(), labels=self.labels, diameter=self.diameter)
        return new_marker

    def key2ind(self, ref):
        ''' Get the index location from a label '''

        if isinstance(ref, (str, unicode)):
            if self.labels is None:
                raise ValueError('Labels must be defined to be used to access markers')
            else:
                return self.labels.index(ref)
        else:
            return int(ref)

    def fromEDM(self, D, labels=None, method='mds'):
        '''
        Compute the position of markers from their Euclidean Distance Matrix

        Parameters
        ----------
        D: square 2D ndarray
            Euclidean Distance Matrix (matrix containing squared distances between points
        labels: list, optional
            A list of human friendly labels for the markers (e.g. 'east', 'west', etc)
        method: str, optional 
            The method to use
            * 'mds' for multidimensional scaling (default)
            * 'tri' for trilateration
        '''

        if D.shape[0] != D.shape[1]:
            raise ValueError('The distance matrix must be square')

        self.m = D.shape[0]

        if method == 'tri':
            self.trilateration(D)
        else:
            self.classical_mds(D)

    def classical_mds(self, D):
        ''' 
        Classical multidimensional scaling

        Parameters
        ----------
        D : square 2D ndarray
            Euclidean Distance Matrix (matrix containing squared distances between points
        '''

        # Apply MDS algorithm for denoising
        n = D.shape[0]
        J = np.eye(n) - np.ones((n,n))/float(n)
        G = -0.5*np.dot(J, np.dot(D, J))

        s, U = np.linalg.eig(G)

        # we need to sort the eigenvalues in decreasing order
        s = np.real(s)
        o = np.argsort(s)
        s = s[o[::-1]]
        U = U[:,o[::-1]]

        S = np.diag(s)[0:self.dim,:]
        self.X = np.dot(np.sqrt(S),U.T)

    def trilateration_single_point(self, c, Dx, Dy):
        '''
        Given x at origin (0,0) and y at (0,c) the distances from a point
        at unknown location Dx, Dy to x, y, respectively, finds the position of the point.
        '''

        z = (c**2 - (Dy**2 - Dx**2)) / (2*c)
        t = np.sqrt(Dx**2 - z**2)

        return np.array([t,z])

    def trilateration(self, D):
        '''
        Find the location of points based on their distance matrix using trilateration

        Parameters
        ----------
        D : square 2D ndarray
            Euclidean Distance Matrix (matrix containing squared distances between points
        '''

        dist = np.sqrt(D)

        # Simpler algorithm (no denoising)
        self.X = np.zeros((self.dim, self.m))

        self.X[:,1] = np.array([0, dist[0,1]])
        for i in range(2,m):
            self.X[:,i] = self.trilateration_single_point(self.X[1,1],
                    dist[0,i], dist[1,i])

    def EDM(self):
        ''' Computes the EDM corresponding to the marker set '''
        if self.X is None:
            raise ValueError('No marker set')

        G = np.dot(self.X.T, self.X)
        return np.outer(np.ones(self.m), np.diag(G)) \
            - 2*G + np.outer(np.diag(G), np.ones(self.m))

    def normalize(self, refs=None):
        '''
        Reposition points such that x0 is at origin, x1 lies on c-axis
        and x2 lies above x-axis, keeping the relative position to each other.
        The z-axis is defined according to right hand rule by default.

        Parameters
        ----------
        refs : list of 3 ints or str
            The index or label of three markers used to define (origin, x-axis, y-axis)
        left_hand : bool, optional (default False)
            Normally the z-axis is defined using right-hand rule, this flag allows to override this behavior
        '''

        if refs is None:
            refs = [0,1,2,3]

        # Transform references to indices if needed
        refs = [self.key2ind(s) for s in refs]

        if self.dim == 2 and len(refs) < 3:
            raise ValueError('In 2D three reference points are needed to define a reference frame')
        elif self.dim == 3 and len(refs) < 4:
            raise ValueError('In 3D four reference points are needed to define a reference frame')

        # set first point to origin
        X0 = self.X[:,refs[0],None]
        Y = self.X - X0

        # Rotate around z to align x-axis to second point
        theta = np.arctan2(Y[1,refs[1]],Y[0,refs[1]])
        c = np.cos(theta)
        s = np.sin(theta)
        if self.dim == 2:
            H = np.array([[c, s],[-s, c]])
        elif self.dim == 3:
            H = np.array([[c, s, 0],[-s, c, 0], [0, 0, 1]])
        Y = np.dot(H,Y)

        if self.dim == 2:
            # set third point to lie above x-axis
            if Y[1,refs[2]] < 0:
                Y[1,:] *= -1

        elif self.dim == 3:
            # In 3D we also want to make sur z-axis points up
            theta = np.arctan2(Y[2,refs[2]],Y[1,refs[2]])
            c = np.cos(theta)
            s = np.sin(theta)
            H = np.array([[1, 0, 0], [0, c, s],[0, -s, c]])
            Y = np.dot(H,Y)

        # Flip the z-axis if requested
        if self.dim == 3 and Y[2,refs[3]] < 0:
            Y[2,:] *= -1

        self.X = Y

    def center(self, marker):
        ''' Translate the marker set so that the argument is the origin. '''

        index = self.key2ind(marker)
        self.X -= self.X[:,index,None]

    def align(self, marker, axis):
        '''
        Rotate the marker set around the given axis until it is aligned onto the given marker

        Parameters
        ----------
        marker : int or str
            the index or label of the marker onto which to align the set
        axis : int
            the axis around which the rotation happens
        '''

        index = self.key2ind(marker)
        axis = ['x','y','z'].index(axis) if isinstance(marker, (str, unicode)) else axis

        # swap the axis around which to rotate to last position
        Y = self.X
        if self.dim == 3:
            Y[axis,:], Y[2,:] = Y[2,:], Y[axis,:]

        # Rotate around z to align x-axis to second point
        theta = np.arctan2(Y[1,index],Y[0,index])
        c = np.cos(theta)
        s = np.sin(theta)
        H = np.array([[c, s],[-s, c]])
        Y[:2,:] = np.dot(H,Y[:2,:])

        if self.dim == 3:
            Y[axis,:], Y[2,:] = Y[2,:], Y[axis,:]

    def flatten(self, ind):
        '''
        Transform the set of points so that the subset of markers given as argument is
        as close as flat (wrt z-axis) as possible.

        Parameters
        ----------
        ind : list of bools
            Lists of marker indices that should be all in the same subspace
        '''

        # Transform references to indices if needed
        ind = [self.key2ind(s) for s in ind]

        # center point cloud around the group of indices
        centroid = self.X[:,ind].mean(axis=1, keepdims=True)
        X_centered = self.X - centroid

        # The rotation is given by left matrix of SVD
        U,S,V = la.svd(X_centered[:,ind], full_matrices=False)

        self.X = np.dot(U.T, X_centered) + centroid

    def correct(self, corr_dic):
        ''' correct a marker location by a given vector '''

        for key, val in corr_dic.items():
            ind = self.key2ind(key)
            self.X[:,ind] += val

    def doa(self, receiver, source):
        ''' Computes the direction of arrival wrt a source and receiver '''

        s_ind = self.key2ind(source)
        r_ind = self.key2ind(receiver)

        # vector from receiver to source
        v = self.X[:,s_ind] - self.X[:,r_ind]

        azimuth = np.arctan2(v[1], v[0])
        elevation = np.arctan2(v[2], la.norm(v[:2]))

        return np.array([azimuth, elevation])

    def plot(self, axes=None, show_labels=True, **kwargs):

        try:
            from mpl_toolkits.mplot3d import Axes3D
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings
            warnings.warn('Matplotlib is required for plotting')
            return

        if self.dim == 2:

            # Create a figure if needed
            if axes is None:
                axes = plt.subplot(111)

            axes.plot(self.X[0,:], self.X[1,:], **kwargs)
            axes.axis(aspect='equal')
            plt.show()


        elif self.dim == 3:
            if axes is None:
                fig = plt.figure()
                axes = fig.add_subplot(111, projection='3d')
            axes.scatter(self.X[0,:], self.X[1,:], self.X[2,:], **kwargs)
             
            axes.set_xlabel('X')
            axes.set_ylabel('Y')
            axes.set_zlabel('Z')
            plt.show()

        if show_labels and self.labels is not None:
            eps = np.linalg.norm(self.X[:,0] - self.X[:,1])/100
            for i in range(self.m):
                if self.dim == 2:
                    axes.text(self.X[0,i]+eps, self.X[1,i]+eps, self.labels[i])
                elif self.dim == 3:
                    axes.text(self.X[0,i]+eps, self.X[1,i]+eps, self.X[2,i]+eps, self.labels[i], None)

        return axes




if __name__ == '__main__':

    import matplotlib.pyplot as plt

    # number of markers
    m = 4
    dim = 2

    D = np.zeros((m,m))

    marker_diameter = 0.040 # 4 cm

    M_orig = MarkerSet(X=np.array([[0.,0.],[0.7,0.],[0.7,0.7],[0.,0.7]]).T)
    D = np.sqrt(M_orig.EDM())

    """
    D[0,1] = D[1,0] = 4.126 + marker_diameter
    D[0,2] = D[2,0] = 6.878 + marker_diameter
    D[0,3] = D[3,0] = 4.508 + marker_diameter
    D[1,2] = D[2,1] = 4.401 + marker_diameter
    D[1,3] = D[3,1] = 7.113 + marker_diameter
    D[3,2] = D[2,3] = 7.002 + marker_diameter
    """

    M1 = MarkerSet(m=m, dim=dim, diameter=marker_diameter)
    M1.fromEDM(D**2)
    M1.normalize()

    M2 = MarkerSet(m=m, dim=dim, diameter=marker_diameter)
    M2.fromEDM(D**2, method='tri')
    M2.normalize()

    M2.plot(marker='ko', labels=True)
    M1.plot(marker='rx')
    plt.show()




# coding=utf-8
# Blind Source Separation for sparsely mixed signals based on Independent Vector Analysis (IVA)
# with Auxiliary Function
# Copyright (C) 2018 Yaron Dibner, Virgile Hernicot, Juan Azcarreta, MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

import numpy as np
from .common import projection_back, sparir


def sparseauxiva(
    X,
    S=None,
    n_src=None,
    n_iter=20,
    proj_back=True,
    W0=None,
    model="laplace",
    return_filters=False,
    callback=None,
):

    """
    Implementation of sparse AuxIVA algorithm for BSS presented in

    J. Janský, Z. Koldovský, and N. Ono, *A computationally cheaper method
    for blind speech separation based on AuxIVA and incomplete demixing transform,*
    Proc. IEEE, IWAENC, pp. 1-5, Sept. 2016.

    Parameters
    ----------
    X: ndarray (nframes, nfrequencies, nchannels)
        STFT representation of the signal
    n_src: int, optional
        The number of sources or independent components
    S: ndarray (k_freq)
        Indexes of active frequency bins for sparse AuxIVA
    n_iter: int, optional
        The number of iterations (default 20)
    proj_back: bool, optional
        Scaling on first mic by back projection (default True)
    W0: ndarray (nfrequencies, nchannels, nchannels), optional
        Initial value for demixing matrix
    model: str
        The model of source distribution 'gauss' or 'laplace' (default)
    return_filters: bool
        If true, the function will return the demixing matrix too
    callback: func
        A callback function called every 10 iterations, allows to monitor convergence

    Returns
    -------
    Returns an (nframes, nfrequencies, nsources) array. Also returns
    the demixing matrix (nfrequencies, nchannels, nsources)
    if ``return_values`` keyword is True.
    """

    n_frames, n_freq, n_chan = X.shape

    assert (
        n_chan == 2
    ), "Current implementation of sparse AuxIVA is specialized for two sources."

    if S is None:
        k_freq = n_freq
        S = np.arange(n_freq)
    else:
        k_freq = S.shape[0]

    # default to determined case
    if n_src is None:
        n_src = X.shape[2]

    # for now it only supports determined case
    assert (
        n_chan == n_src
    ), "Only the determined case is implemented (n_src == n_channels)."

    if model not in ["laplace", "gauss"]:
        raise ValueError("Model should be either " "laplace" " or " "gauss" ".")

    # initialize the demixing matrices
    if W0 is None:
        W = np.array([np.eye(n_chan, n_src) for f in range(n_freq)], dtype=X.dtype)
    else:
        W = W0.copy()

    eps = 1e-10
    I = np.eye(n_src, n_src)
    Y = np.zeros((n_frames, n_freq, n_src), dtype=X.dtype)
    r = np.zeros((n_frames, n_src))

    def demixsparse(Y, X, S, W):
        for f in range(k_freq):
            Y[:, S[f], :] = np.dot(X[:, S[f], :], np.conj(W[S[f], :, :]))

    # Conventional AuxIVA in the frequency bins k_freq selected by S
    for epoch in range(n_iter):
        demixsparse(Y, X, S, W)

        if callback is not None and epoch % 10 == 0:
            if proj_back:
                z = projection_back(Y, X[:, :, 0])
                callback(Y * np.conj(z[None, :, :]))
            else:
                callback(Y)

        # shape: (n_frames, n_src)
        if model == "laplace":
            r[:, :] = 2.0 * np.linalg.norm(Y, axis=1)
        elif model == "gauss":
            r[:, :] = (np.linalg.norm(Y, axis=1) ** 2) / n_freq

        # ensure some numerical stability
        r[r < eps] = eps

        r_inv = 1.0 / r

        # Compute Auxiliary Variable
        V = np.mean(
            (X[:, :, None, :, None] * r_inv[:, None, :, None, None])
            * np.conj(X[:, :, None, None, :]),
            axis=0,
        )

        # Update now the demixing matrix
        for s in range(n_src):
            W_H = np.conj(np.swapaxes(W, 1, 2))
            WV = np.matmul(W_H, V[:, s, :, :])
            rhs = I[None, :, s][[0] * WV.shape[0], :]
            W[:, :, s] = np.linalg.solve(WV, rhs)

            # normalize
            P1 = np.conj(W[:, :, s])
            P2 = np.sum(V[:, s, :, :] * W[:, None, :, s], axis=-1)
            W[:, :, s] /= np.sqrt(np.sum(P1 * P2, axis=1))[:, None]

    # LASSO regularization to reconstruct the complete relative transfer function
    Z = np.zeros((n_src, k_freq), dtype=W.dtype)
    G = np.zeros((n_src, n_freq, 1), dtype=Z.dtype)
    hrtf = np.zeros((n_freq, n_src), dtype=W.dtype)

    for i in range(n_src):
        # calculate sparse relative transfer function from demixing matrix
        Z[i, :] = np.conj(-W[S, 0, i] / W[S, 1, i]).T

        # copy selected active frequencies in Z to sparse G
        G[i, S] = np.expand_dims(Z[i, :], axis=1)

        # apply fast proximal algorithm to reconstruct the complete real-valued relative transfer function
        hrtf[:, i] = sparir(G[i, :], S)

        # recover relative transfer function back to the frequency domain
        hrtf[:, i] = np.fft.fft(hrtf[:, i])
        # assemble back the complete demixing matrix
        W[:, :, i] = np.conj(np.insert(hrtf[:, i, None], 1, -1, axis=1))

    # apply final demixing in the whole frequency range
    demixsparse(Y, X, np.array(range(n_freq)), W)

    # Projection back to correct scale ambiguity
    if proj_back:
        z = projection_back(Y, X[:, :, 0])
        Y *= np.conj(z[None, :, :])

    if return_filters:
        return Y, W
    else:
        return Y
